import pandas as pd

from pyvest import FamaFrenchDataReader, YFDataReader


class Asset:
    def __init__(self, ticker):
        self.__ticker = ticker

    @property
    def ticker(self):
        return self.__ticker

    @property
    def start_date(self):
        raise NotImplementedError("Asset.start_date has not been implemented.")

    @property
    def end_date(self):
        raise NotImplementedError("Asset.start_date has not been implemented.")

    @property
    def dates(self):
        raise NotImplementedError("Asset.dates has not been implemented.")

    def get_price(self, date):
        raise NotImplementedError("Asset.get_price has not been implemented.")

    def __repr__(self):
        return self.__generate_output()

    def __str__(self):
        return self.__generate_output()

    def __generate_output(self):
        output = "Ticker: " + self.__ticker

        return output


class Stock(Asset):
    def __init__(self, ticker, start_date=None, end_date=None,
                 frequency="1d", adjusted_close=False, data_reader=None):
        super().__init__(ticker)

        self.__start_date = start_date
        self.__end_date = end_date
        self.__frequency = frequency
        self.__adjusted_close = adjusted_close

        self.__data_reader = YFDataReader() if data_reader is None \
            else data_reader

        self.__history_df = None

        if self.__data_reader is not None and self.__start_date is not None \
                and self.__end_date is not None:
            self.__read_history()
            self.__extract_dividends_from_history()

    def __repr__(self):
        return super().__repr__() + "\n" + self.__generate_output()

    def __str__(self):
        return super().__str__() + "\n" + self.__generate_output()

    @property
    def start_date(self):
        return self.__history_df.index[0]

    @property
    def end_date(self):
        return self.__history_df.index[-1]

    @property
    def dates(self):
        return self.__history_df.index

    @property
    def dividends(self):
        return self.__dividends_series

    def get_price(self, date):
        column = "Adj Close" if self.__adjusted_close else "Close"
        asset_price = self.__history_df.loc[date][column]

        return asset_price

    def __read_history(self):
        self.__history_df = self.__data_reader.read_history(
            self.ticker, self.__start_date, self.__end_date,
            interval=self.__frequency)
        self.__history_df.index = \
            self.__history_df.index.map(lambda x: x.replace(tzinfo=None))

    def __extract_dividends_from_history(self):
        self.__dividends_series = self.__history_df["Dividends"][
            self.__history_df["Dividends"] > 0]

        return self.__dividends_series

    def __generate_output(self):
        output = ""

        return output


class StockPortfolio(Asset):
    def __init__(self, ticker, start_date=None, end_date=None,
                 frequency="monthly", initial_price=1.0, data_reader=None):

        super().__init__(ticker)

        self.__start_date = start_date
        self.__end_date = end_date
        self.__frequency = frequency
        self.__initial_price = initial_price

        self.__data_reader = FamaFrenchDataReader() if data_reader is None \
            else data_reader

        self.__history_df = None

        if self.__data_reader is not None and self.__start_date is not None \
                and self.__end_date is not None:
            self.__read_returns()
            self.__generate_price_history()

    def __repr__(self):
        return super().__repr__() + "\n" + self.__generate_output()

    def __str__(self):
        return super().__str__() + "\n" + self.__generate_output()

    @property
    def start_date(self):
        return self.__history_df.index[0]

    @property
    def end_date(self):
        return self.__history_df.index[-1]

    @property
    def dates(self):
        return self.__history_df.index

    def get_price(self, date):
        asset_price = self.__history_df.loc[date]["price"]

        return asset_price

    def __read_returns(self):

        if self.__is_factor(self.ticker):
            self.__read_factor()
        else:
            self.__read_portfolio()

    def __is_factor(self, ticker):

        is_factor = False

        factors_tickers = ["Mkt-RF", "SMB", "HML", "RF", "Mkt"]
        if ticker in factors_tickers:
            is_factor = True

        return is_factor

    def __read_factor(self):

        factors_df = self.__data_reader.read_factors(self.__start_date,
                                                     self.__end_date)

        if self.ticker == "Mkt":
            self.__history_df = pd.DataFrame(factors_df["Mkt-RF"]
                                             + factors_df["RF"],
                                             columns=["return"])
        else:
            column_name = self.ticker
            self.__history_df = factors_df[[column_name]]
            self.__history_df = self.__history_df.rename(
                columns={column_name: "return"})

        self.__history_df.index = self.__history_df.index.to_timestamp()

    def __read_portfolio(self):

        data_set, portfolio_index = self.__get_portfolio_name_from_ticker()
        portfolios_df = self.__data_reader.read_returns(data_set,
                                                        self.__start_date,
                                                        self.__end_date)
        column_name = portfolio_index
        self.__history_df = portfolios_df[[column_name]]
        self.__history_df = self.__history_df.rename(
            columns={column_name: "return"})
        self.__history_df.index = self.__history_df.index.to_timestamp()

    def __get_portfolio_name_from_ticker(self):

        portfolio_mapping_dict = {
            "beta": 'Portfolios_Formed_on_BETA',
            "size": 'Portfolios_Formed_on_ME',
            "value": 'Portfolios_Formed_on_BE-ME',
            "momentum": '10_Portfolios_Prior_12_2'
        }

        data_set = portfolio_mapping_dict[self.ticker.split("_")[0]]

        portfolio_index = self.ticker.split("_")[1]

        return data_set, portfolio_index

    def __generate_price_history(self):
        returns_series = self.__history_df["return"]
        # The return of the first month is omitted since the first price should
        # correspond to the last day of the first month.
        returns_series.iloc[0] = 0.0
        prices_series = (1.0 + returns_series * 0.01).cumprod()

        self.__history_df["price"] = prices_series

    def __generate_output(self):
        output = ""

        return output



class Profit:
    def __init__(self, investment, date=None,
                 realized_capital_gain_by_ticker=None,
                 unrealized_capital_gain_by_ticker=None,
                 reinvested_dividends_by_ticker=None,
                 non_reinvested_dividends_by_ticker=None,
                 profit_to_copy=None):

        self.__investment = investment

        self.__date = date

        self.__init_values_by_tickers()

        self.__realized_capital_gain = None
        self.__unrealized_capital_gain = None
        self.__reinvested_dividends = None
        self.__non_reinvested_dividends = None

        if profit_to_copy is not None:
            self.__copy(profit_to_copy)

        self.__initialize_attributes(date, realized_capital_gain_by_ticker,
                                     unrealized_capital_gain_by_ticker,
                                     reinvested_dividends_by_ticker,
                                     non_reinvested_dividends_by_ticker)

        self.__calculate_investment_profit()

    def __repr__(self):
        return self.__generate_output()

    def __str__(self):
        return self.__generate_output()

    @property
    def date(self):
        return self.__date

    @property
    def realized_capital_gain_by_ticker(self):
        return self.__realized_capital_gain_by_ticker

    @property
    def unrealized_capital_gain_by_ticker(self):
        return self.__unrealized_capital_gain_by_ticker

    @property
    def reinvested_dividends_by_ticker(self):
        return self.__reinvested_dividends_by_ticker

    @property
    def non_reinvested_dividends_by_ticker(self):
        return self.__non_reinvested_dividends_by_ticker

    @property
    def realized_capital_gain(self):
        return self.__realized_capital_gain

    @property
    def unrealized_capital_gain(self):
        return self.__unrealized_capital_gain

    @property
    def reinvested_dividends(self):
        return self.__reinvested_dividends

    @property
    def non_reinvested_dividends(self):
        return self.__non_reinvested_dividends

    def __init_values_by_tickers(self):

        self.__realized_capital_gain_by_ticker = {}
        self.__unrealized_capital_gain_by_ticker = {}
        self.__reinvested_dividends_by_ticker = {}
        self.__non_reinvested_dividends_by_ticker = {}

        for ticker in self.__investment.assets.keys():
            self.__realized_capital_gain_by_ticker[ticker] = 0
            self.__unrealized_capital_gain_by_ticker[ticker] = 0
            self.__reinvested_dividends_by_ticker[ticker] = 0
            self.__non_reinvested_dividends_by_ticker[ticker] = 0

    def __copy(self, profit_to_copy):
        self.__date = profit_to_copy.date
        self.__realized_capital_gain_by_ticker = \
            profit_to_copy.realized_capital_gain_by_ticker.copy()
        self.__unrealized_capital_gain_by_ticker = \
            profit_to_copy.unrealized_capital_gain_by_ticker.copy()
        self.__reinvested_dividends_by_ticker = \
            profit_to_copy.reinvested_dividends_by_ticker.copy()
        self.__non_reinvested_dividends_by_ticker = \
            profit_to_copy.non_reinvested_dividends_by_ticker.copy()

    def __initialize_attributes(self, date, realized_capital_gain_by_ticker,
                                unrealized_capital_gain_by_ticker,
                                reinvested_dividends_by_ticker,
                                non_reinvested_dividends_by_ticker):

        if date is not None:
            self.__date = date

        if realized_capital_gain_by_ticker is not None:
            for ticker, value in realized_capital_gain_by_ticker.items():
                self.__realized_capital_gain_by_ticker[ticker] = value

        if unrealized_capital_gain_by_ticker is not None:
            for ticker, value in unrealized_capital_gain_by_ticker.items():
                self.__unrealized_capital_gain_by_ticker[ticker] = value

        if reinvested_dividends_by_ticker is not None:
            for ticker, value in reinvested_dividends_by_ticker.items():
                self.__reinvested_dividends_by_ticker[ticker] = value

        if non_reinvested_dividends_by_ticker is not None:
            for ticker, value in non_reinvested_dividends_by_ticker.items():
                self.__non_reinvested_dividends_by_ticker[ticker] = value

    def __calculate_investment_profit(self):

        self.__realized_capital_gain = 0
        for ticker, value in self.__realized_capital_gain_by_ticker.items():
            self.__realized_capital_gain += value

        self.__unrealized_capital_gain = 0
        for ticker, value in self.__unrealized_capital_gain_by_ticker.items():
            self.__unrealized_capital_gain += value

        self.__reinvested_dividends = 0
        for ticker, value in self.__reinvested_dividends_by_ticker.items():
            self.__reinvested_dividends += value

        self.__non_reinvested_dividends = 0
        for ticker, value in self.__non_reinvested_dividends_by_ticker.items():
            self.__non_reinvested_dividends += value

    def __generate_output(self):
        output = ""

        if self.__date is not None:
            output += "Date: " + str(self.__date)

        if self.__realized_capital_gain is not None:
            if len(output) > 0:
                output += "\n"
            output += "Realized Capital Gain: " \
                      + str(self.__realized_capital_gain)

        if self.__unrealized_capital_gain is not None:
            if len(output) > 0:
                output += "\n"
            output += "Unrealized Capital Gain: " \
                      + str(self.__unrealized_capital_gain)

        if self.__reinvested_dividends is not None:
            if len(output) > 0:
                output += "\n"
            output += "Reinvested Dividends: " \
                      + str(self.__reinvested_dividends)

        if self.__non_reinvested_dividends is not None:
            if len(output) > 0:
                output += "\n"
            output += "Non Reinvested Dividends: " \
                      + str(self.__non_reinvested_dividends)

        return output



import numpy as np
import matplotlib.pyplot as plt


###############################################################################
######################################### SIMULATION ##########################
###############################################################################

class Simulation:

    def __init__(self, portfolio, initial_value, periodic_contribution,
                 nb_periods):
        self.__portfolio = portfolio
        self.__initial_value = initial_value
        self.__periodic_contribution = periodic_contribution
        self.__nb_periods = nb_periods

        self.__portfolio_value_paths_list = None
        self.__portfolio_value_mean_path = None
        self.__portfolio_value_std_path = None

    def calculate_portfolio_value_paths(self, nb_paths, random_seed=1):

        np.random.seed(
            seed=random_seed)  # Fix the seed of the random number generator.
        self.__portfolio_value_paths_list = []
        # Initiate empty list to contain the paths of the simulation

        for i in range(0, nb_paths):
            r_portfolio_array = np.random.normal(
                self.__portfolio.expected_return,
                self.__portfolio.standard_deviation,
                self.__nb_periods)  # Draw realizations of the market return
            r_portfolio_array = r_portfolio_array.clip(
                min=-1)  # We consider a loss greater than 100% as a 100% loss.

            portfolio_value_path = [
                self.__initial_value]  # Initiate sample path to "initial_value"
            current_value = self.__initial_value
            for annual_return in r_portfolio_array:
                current_value = current_value * (
                        1 + annual_return) + self.__periodic_contribution
                # Cumulate return and contribution
                portfolio_value_path.append(
                    current_value)  # Append current wealth to path

            self.__portfolio_value_paths_list.append(
                portfolio_value_path)  # Append current path to simulation

        return self.__portfolio_value_paths_list

    ########################## read-only attributes ###########################
    @property
    def portfolio_value_paths(self):
        return self.__portfolio_value_paths_list

    @property
    def portfolio_value_mean_path(self):
        if self.__portfolio_value_paths_list:
            return list(np.mean(self.__portfolio_value_paths_list, axis=0))
        else:
            return None

    @property
    def portfolio_value_std_path(self):
        if self.__portfolio_value_paths_list:
            return list(
                np.std(self.__portfolio_value_paths_list, axis=0, ddof=1))
        else:
            return None

    @property
    def portfolio_final_values(self):
        if self.__portfolio_value_paths_list:
            return [portfolio_value_array[-1] for portfolio_value_array in
                    self.__portfolio_value_paths_list]
        else:
            return None


###############################################################################
################################# SimulationVisualizer ########################
###############################################################################

class SimulationVisualizer:

    def __init__(self, simulations, labels=None,
                 colors=None):

        if colors is None:
            colors = ['blue', 'red', 'green', 'yellow']
        if labels is None:
            labels = ["1", "2", "3", "4"]
        if isinstance(simulations, Simulation):
            self.__simulations = [simulations]
        else:
            self.__simulations = simulations
        self.__labels = labels
        self.__colors = colors

        self.__fig = None
        self.__ax = None
        self.__fig_twd = None
        self.__ax_twd = None

    ################################# ATTRIBUTES ##############################

    @property
    def fig(self):
        return self.__fig

    @property
    def ax(self):
        return self.__ax

    @property
    def fig_twd(self):
        return self.__fig_twd

    @property
    def ax_twd(self):
        return self.__ax_twd

    def show_portfolios_value(self, nb_paths_max=100):

        self.__fig, self.__ax = plt.subplots(figsize=(16, 10))
        self.__ax.set_title('Portfolio value')
        self.__ax.set_xlabel('Number of years')
        self.__ax.set_ylabel('Portfolio value')

        colors_iter = iter(self.__colors)
        labels_iter = iter(self.__labels)
        for simulation in self.__simulations:
            color = next(colors_iter)
            label = next(labels_iter)
            nb_paths = 0
            for portfolio_value_path in simulation.portfolio_value_paths:
                self.__ax.plot(portfolio_value_path, color=color, alpha=0.2,
                               linewidth=1)
                nb_paths += 1
                if nb_paths >= nb_paths_max:
                    break;

        colors_iter = iter(self.__colors)
        labels_iter = iter(self.__labels)
        for simulation in self.__simulations:
            color = next(colors_iter)
            label = next(labels_iter)
            self.__ax.plot(simulation.portfolio_value_mean_path, color=color,
                           linewidth=3, label=label)

        self.__ax.legend()

    def show_terminal_wealth_distribution(self):

        self.__fig_twd, self.__ax_twd = plt.subplots(figsize=(16, 10))
        self.__ax_twd.set_title('Distribution of terminal wealth')
        self.__ax_twd.set_xlabel('Terminal wealth')
        self.__ax_twd.set_ylabel('Number of paths')

        colors_iter = iter(self.__colors)
        labels_iter = iter(self.__labels)
        for simulation in self.__simulations:
            color = next(colors_iter)
            label = next(labels_iter)
            self.__ax_twd.hist(simulation.portfolio_final_values, color=color,
                               alpha=0.5, bins=100)
            self.__ax_twd.axvline(np.mean(simulation.portfolio_final_values),
                                  color=color, label=label)

        self.__ax_twd.legend()



class ReturnsDataReader:
    def __init__(self):
        pass

    def read_returns(self, assets, start_date, end_date):
        raise NotImplementedError("ReturnsDataReader.read_returns has not "
                                  "been implemented.")



import matplotlib.pyplot as plt
import matplotlib
import numpy as np

import pyvest.investment_universe.investment_universe \
    as investment_universe_module
import pyvest.general


class InvestmentUniverseVisualizer:
    MAX_NB_INV_UNIV = 4
    LINE_NB_POINTS = 100
    FILL_BETWEEN_X_EXCESS = 0.1

    class VisualElement:
        def __init__(self, plot_function, priority, size,
                     investment_universe=None, label=None, investors=None,
                     indifference_curves=None, nb_curves=None):
            self.__plot_function = plot_function
            self.__investment_universe = investment_universe
            self.__priority = priority
            self.__size = size
            self.__label = label
            self.__investors = investors
            self.__indifference_curves = indifference_curves
            self.__nb_curves = nb_curves

            self.__zorder = 1 / self.__priority

        def plot(self):
            if self.__investment_universe is not None \
                    and (self.__investors is not None
                         or self.__indifference_curves is not None) \
                    and self.__nb_curves is not None:
                self.__plot_function(self.__investment_universe,
                                     self.__investors,
                                     self.__indifference_curves, self.__label,
                                     self.__size, self.__zorder,
                                     self.__nb_curves)
            elif self.__investment_universe is not None \
                    and (self.__investors is not None
                         or self.__indifference_curves is not None):
                self.__plot_function(self.__investment_universe,
                                     self.__investors,
                                     self.__indifference_curves, self.__label,
                                     self.__size, self.__zorder)
            elif self.__investment_universe is not None:
                self.__plot_function(self.__investment_universe, self.__label,
                                     self.__size, self.__zorder)
            else:
                self.__plot_function(self.__label, self.__size, self.__zorder)

        def __lt__(self, other):
            return self.__priority < other.__priority

    def __init__(self, investment_universes, labels=None,
                 default_visibility=True, weights_visible=True,
                 nb_decimal_places=2):

        self.__assign_investment_universes(investment_universes)
        self.__assign_labels(labels)
        self.__generate_assets_inv_univ_dict()

        self.__alpha = 1.0

        self.__ax = None
        self.__fig = None

        self.__set_default_visibility(default_visibility)
        self.__weights_visible = weights_visible
        self.__nb_decimal_places = nb_decimal_places

        self.reset_colors()
        self.__set_default_visual_elements_properties()

        self.__visual_elements_list = None
        self.__investors_colors = {}

    ################################# ATTRIBUTES #################################

    @property
    def assets_visible(self):
        return self.__assets_visible

    @assets_visible.setter
    def assets_visible(self, value):
        self.__check_bool(value, "assets_visible")
        self.__assets_visible = value

    @property
    def feasible_portfolios_with_r_f_visible(self):
        return self.__feasible_portfolios_with_r_f_visible

    @feasible_portfolios_with_r_f_visible.setter
    def feasible_portfolios_with_r_f_visible(self, value):
        self.__check_bool(value, "feasible_portfolios_with_r_f_visible")
        self.__feasible_portfolios_with_r_f_visible = value

    @property
    def feasible_portfolios_visible(self):
        return self.__feasible_portfolios_visible

    @feasible_portfolios_visible.setter
    def feasible_portfolios_visible(self, value):
        self.__check_bool(value, "feasible_portfolios_visible")
        self.__feasible_portfolios_visible = value

    @property
    def mvp_visible(self):
        return self.__mvp_visible

    @mvp_visible.setter
    def mvp_visible(self, value):
        self.__check_bool(value, "mvp_visible")
        self.__mvp_visible = value

    @property
    def efficient_frontier_visible(self):
        return self.__efficient_frontier_visible

    @efficient_frontier_visible.setter
    def efficient_frontier_visible(self, value):
        self.__check_bool(value, "efficient_frontier_visible")
        self.__efficient_frontier_visible = value

    @property
    def tangency_portfolio_visible(self):
        return self.__tangency_portfolio_visible

    @tangency_portfolio_visible.setter
    def tangency_portfolio_visible(self, value):
        self.__check_bool(value, "tangency_portfolio_visible")
        self.__tangency_portfolio_visible = value

    @property
    def cal_visible(self):
        return self.__cal_visible

    @cal_visible.setter
    def cal_visible(self, value):
        self.__check_bool(value, "cal_visible")
        self.__cal_visible = value

    @property
    def r_f_visible(self):
        return self.__r_f_visible

    @r_f_visible.setter
    def r_f_visible(self, value):
        self.__check_bool(value, "r_f_visible")
        self.__r_f_visible = value

    @property
    def other_portfolios_visible(self):
        return self.__other_portfolios_visible

    @other_portfolios_visible.setter
    def other_portfolios_visible(self, value):
        self.__check_bool(value, "other_portfolios_visible")
        self.__other_portfolios_visible = value

    @property
    def weights_visible(self):
        return self.__weights_visible

    @weights_visible.setter
    def weights_visible(self, value):
        self.__weights_visible = value

    @property
    def market_portfolio_visible(self):
        return self.__market_portfolio_visible

    @market_portfolio_visible.setter
    def market_portfolio_visible(self, value):
        self.__market_portfolio_visible = value

    @property
    def nb_decimal_places(self):
        return self.__nb_decimal_places

    @nb_decimal_places.setter
    def nb_decimal_places(self, value):
        self.__nb_decimal_places = value

    @property
    def investors(self):
        return self.__investors

    @investors.setter
    def investors(self, value):
        self.__check_bool(value, "investors")
        self.__investors = value

    @property
    def min_expected_return(self):
        return self.__min_expected_return

    @min_expected_return.setter
    def min_expected_return(self, value):
        self.__min_expected_return = value

    @property
    def max_expected_return(self):
        return self.__max_expected_return

    @max_expected_return.setter
    def max_expected_return(self, value):
        self.__max_expected_return = value

    @property
    def min_standard_deviation(self):
        return self.__min_standard_deviation

    @min_standard_deviation.setter
    def min_standard_deviation(self, value):
        self.__min_standard_deviation = value

    @property
    def max_standard_deviation(self):
        return self.__max_standard_deviation

    @max_standard_deviation.setter
    def max_standard_deviation(self, value):
        self.__max_standard_deviation = value

    @property
    def investment_universes(self):
        return self.__investment_universes

    @investment_universes.setter
    def investment_universes(self, value):
        self.__assign_investment_universes(value)
        self.__assign_labels(None)

    @property
    def labels(self):
        return self.__labels

    @labels.setter
    def labels(self, value):
        self.__assign_labels(value)

    @property
    def visibility_priorities(self):
        return self.__visibility_priorities

    @visibility_priorities.setter
    def visibility_priorities(self, value):
        self.__visibility_priorities = value

    ##################### fig ###################
    @property
    def fig(self):
        return self.__fig

    ##################### ax ###################
    @property
    def ax(self):
        return self.__ax

    ##################### colors ###################
    @property
    def colors(self):
        return self.__colors

    ##################### visual_elements_properties ###################
    @property
    def visual_elements_properties(self):
        return self.__visual_elements_properties

    ######################### PUBLIC FUNCTIONS ########################

    def plot(self, figsize=(16, 9), zoom_individual=False,
             min_expected_return=None, max_expected_return=None,
             min_standard_deviation=None, max_standard_deviation=None,
             investors=None, indifference_curves=None, optimal_portfolios=True,
             legend='upper left'):

        self.__calculate_visible_portfolios_mu_std(zoom_individual, investors)
        self.__set_default_std_limits()
        self.__set_default_mu_limits()

        if min_expected_return is not None:
            self.__min_expected_return = min_expected_return
        if max_expected_return is not None:
            self.__max_expected_return = max_expected_return
        if min_standard_deviation is not None:
            self.__min_standard_deviation = min_standard_deviation
        if max_standard_deviation is not None:
            self.__max_standard_deviation = max_standard_deviation

        self.__fig, self.__ax = plt.subplots(figsize=figsize)

        self.__ax.set_xlim([self.__min_standard_deviation,
                            self.__max_standard_deviation])
        self.__ax.set_ylim([self.__min_expected_return,
                            self.__max_expected_return])
        self.__ax.grid()

        self.__ax.set_title("Risk-return tradeoff", fontsize=35)
        self.__ax.set_ylabel("Expected return", fontsize=30)
        self.__ax.set_xlabel("Standard deviation", fontsize=30)
        self.__ax.tick_params(axis='both', labelsize=25)

        self.__generate_visual_elements_list(investors, indifference_curves,
                                             optimal_portfolios)

        sorted_visual_elements = sorted(self.__visual_elements_list,
                                        reverse=True)
        for vis_elem in sorted_visual_elements:
            vis_elem.plot()

        if type(legend) is str:
            self.__ax.legend(fontsize=15, loc=legend)

    ########################## PRIVATE ##########################

    def __set_default_visibility(self, default_visibility):
        self.__assets_visible = default_visibility
        self.__feasible_portfolios_visible = default_visibility
        self.__feasible_portfolios_with_r_f_visible = default_visibility
        self.__mvp_visible = default_visibility
        self.__efficient_frontier_visible = default_visibility
        self.__tangency_portfolio_visible = default_visibility
        self.__cal_visible = default_visibility
        self.__r_f_visible = default_visibility
        self.__other_portfolios_visible = default_visibility
        self.__market_portfolio_visible = default_visibility

    def __set_default_mu_limits(self, border_padding=0.1):

        mu_list = [mu for mu, std in self.__visible_portfolios_mu_std_list]
        border_abs_value = border_padding * max(mu_list)

        self.__min_expected_return = 0 if min(mu_list) >= 0 \
            else min(mu_list) - border_abs_value
        self.__max_expected_return = max(mu_list) + border_abs_value

    def __set_default_std_limits(self, border_padding=0.1):

        std_list = [std for mu, std in self.__visible_portfolios_mu_std_list]
        border_abs_value = border_padding * max(std_list)

        self.__min_standard_deviation = 0 if min(std_list) >= 0 \
            else min(std_list) - border_abs_value
        self.__max_standard_deviation = max(std_list) + border_abs_value

    def __calculate_visible_portfolios_mu_std(self, zoom_individual,
                                              investors):
        self.__visible_portfolios_mu_std_list = []

        remaining_ptfs_list = []
        for inv_uni in self.__investment_universes:
            self.__visible_portfolios_mu_std_list.extend(
                list(zip(inv_uni.mu, inv_uni.std)))
            if inv_uni.mvp is not None:
                remaining_ptfs_list.append(inv_uni.mvp)
            if inv_uni.tangency_portfolio is not None:
                remaining_ptfs_list.append(inv_uni.tangency_portfolio)
            if inv_uni.other_portfolios is not None:
                other_portfolios = [ptf_name_pair[0] for ptf_name_pair
                                    in list(inv_uni.other_portfolios.values())]
                remaining_ptfs_list.extend(other_portfolios)
            if inv_uni.investors is not None:
                if investors is None:
                    investors = list(inv_uni.investors.keys())
                investor_portfolios = [inv_uni.investors[x].portfolio
                                       for x in investors
                                       if inv_uni.investors[x].portfolio
                                       is not None]
                remaining_ptfs_list.extend(investor_portfolios)

            if not zoom_individual \
                    and inv_uni.efficient_frontier is not None:
                remaining_ptfs_list.extend(inv_uni.efficient_frontier)
            if not zoom_individual \
                    and inv_uni.feasible_portfolios is not None:
                remaining_ptfs_list.extend(inv_uni.feasible_portfolios)

        self.__visible_portfolios_mu_std_list.extend(
            [(ptf.expected_return, ptf.standard_deviation) for ptf
             in remaining_ptfs_list])

    def __assign_investment_universes(self, investment_universes):
        if isinstance(investment_universes,
                      investment_universe_module.InvestmentUniverse):
            self.__investment_universes = [investment_universes]
        else:
            self.__investment_universes = investment_universes

    def __assign_labels(self, labels):
        generic_labels = ["1", "2", "3", "4"]
        if labels is None and len(self.__investment_universes) > 1:
            self.__labels = generic_labels
        elif labels is None:
            self.__labels = []
        else:
            self.__labels = labels + generic_labels[len(labels):]

    def __generate_assets_inv_univ_dict(self):
        self.__assets_inv_univ_dict = {}
        inv_univ_index = 0
        for investment_universe in self.__investment_universes:
            asset_index = 0
            for asset in investment_universe.assets:
                if asset not in self.__assets_inv_univ_dict:
                    self.__assets_inv_univ_dict[asset] = (inv_univ_index,
                                                          asset_index)
                asset_index += 1
            inv_univ_index += 1

    def __plot_feasible_portfolios(self, investment_universe, label, size,
                                   zorder):
        color_label = label if label is not None else "1"
        color = self.__colors[color_label]["feasible"]

        feasible_portfolios_mu_list = list(map(lambda x: x.expected_return,
                                               investment_universe.feasible_portfolios))
        feasible_portfolios_std_list = list(map(lambda x: x.standard_deviation,
                                                investment_universe.feasible_portfolios))

        legend_label = self.__complete_label("Feasible portfolios", label)
        self.__ax.scatter(feasible_portfolios_std_list,
                          feasible_portfolios_mu_list,
                          s=size,
                          alpha=self.__alpha,
                          label=legend_label,
                          color=color, zorder=zorder)

    def __plot_feasible_portfolios_with_r_f(self, investment_universe, label,
                                            size, zorder):
        color_label = label if label is not None else "1"
        color = self.__colors[color_label]["feasible_with_r_f"]

        feasible_portfolios_mu_list = \
            list(map(lambda x: x.expected_return,
                     investment_universe.feasible_portfolios_with_r_f))
        feasible_portfolios_std_list = \
            list(map(lambda x: x.standard_deviation,
                     investment_universe.feasible_portfolios_with_r_f))

        legend_label = self.__complete_label("Feasible portfolios with r_f",
                                             label)
        self.__ax.scatter(feasible_portfolios_std_list,
                          feasible_portfolios_mu_list,
                          s=size,
                          alpha=self.__alpha,
                          label=legend_label,
                          color=color, zorder=zorder)

    def __plot_feasible_portfolios_surface(self, investment_universe, label,
                                           size, zorder):
        color_label = label if label is not None else "1"
        color = self.__colors[color_label]["feasible"]

        mu_sigma_tuples_list = []
        for ptf_pair in investment_universe.feasible_portfolios_surface:
            mu = ptf_pair[0].expected_return
            sigma_min = ptf_pair[0].standard_deviation
            sigma_max = ptf_pair[1].standard_deviation

            mu_sigma_tuples_list.append((mu, sigma_min, sigma_max))

        mu_sigma_tuples_list = sorted(mu_sigma_tuples_list, key=lambda x: x[0])
        mu_sigma_min_max_list = list(zip(*mu_sigma_tuples_list))
        feasible_portfolios_mu_list = mu_sigma_min_max_list[0]
        feasible_portfolios_std_min_list = mu_sigma_min_max_list[1]
        feasible_portfolios_std_max_list = mu_sigma_min_max_list[2]

        legend_label = self.__complete_label("Feasible portfolios", label)

        self.__ax.fill_betweenx(feasible_portfolios_mu_list,
                                feasible_portfolios_std_min_list,
                                feasible_portfolios_std_max_list,
                                linewidth=size,
                                alpha=self.__alpha,
                                label=legend_label,
                                color=color, zorder=zorder)

    def __plot_feasible_portfolios_surface_with_r_f(self, investment_universe,
                                                    label, size, zorder):
        color_label = label if label is not None else "1"
        color = self.__colors[color_label]["feasible_with_r_f"]

        mu_sigma_tuples_list = []
        for ptf_pair in investment_universe.feasible_portfolios_surface_with_r_f:
            mu = ptf_pair[0].expected_return
            sigma_min = ptf_pair[0].standard_deviation
            sigma_max = ptf_pair[1].standard_deviation

            mu_sigma_tuples_list.append((mu, sigma_min, sigma_max))

        mu_sigma_tuples_list = sorted(mu_sigma_tuples_list, key=lambda x: x[0])
        mu_sigma_min_max_list = list(zip(*mu_sigma_tuples_list))
        feasible_portfolios_mu_list = mu_sigma_min_max_list[0]
        feasible_portfolios_std_min_list = mu_sigma_min_max_list[1]
        feasible_portfolios_std_max_list = mu_sigma_min_max_list[2]

        legend_label = self.__complete_label("Feasible portfolios with r_f",
                                             label)

        self.__ax.fill_betweenx(feasible_portfolios_mu_list,
                                feasible_portfolios_std_min_list,
                                feasible_portfolios_std_max_list,
                                linewidth=size,
                                alpha=self.__alpha,
                                label=legend_label,
                                color=color, zorder=zorder)

    def __plot_feasible_portfolios_equation(self, investment_universe, label,
                                            size, zorder):
        color_label = label if label is not None else "1"
        color = self.__colors[color_label]["feasible"]
        legend_label = self.__complete_label("Feasible portfolios", label)

        if len(investment_universe.assets) > 1:
            nb_points = self.LINE_NB_POINTS
            step = (self.__max_expected_return
                    - self.__min_expected_return) / nb_points
            exp_ret_list = list(
                np.arange(self.__min_expected_return,
                          self.__max_expected_return,
                          step))
            std_list = [investment_universe.feasible_portfolios_equation(mu)
                        for mu in exp_ret_list]
        else:
            exp_ret_list = [investment_universe.mu[0]]
            std_list = [investment_universe.std[0]]

        if len(investment_universe.assets) < 3:
            self.__ax.plot(std_list, exp_ret_list, linewidth=size,
                           alpha=self.__alpha, label=legend_label, color=color,
                           zorder=zorder)
        else:
            max_std = (1 + self.FILL_BETWEEN_X_EXCESS) \
                      * self.__max_standard_deviation
            self.__ax.fill_betweenx(
                exp_ret_list, std_list,
                max_std * self.__max_standard_deviation,
                linewidth=size, alpha=self.__alpha, label=legend_label,
                color=color, zorder=zorder)

    def __plot_feasible_portfolios_equation_with_r_f(self, investment_universe,
                                                     label, size, zorder):
        color_label = label if label is not None else "1"
        color = self.__colors[color_label]["feasible_with_r_f"]

        nb_points = self.LINE_NB_POINTS
        step = (
                       self.__max_expected_return - self.__min_expected_return) / nb_points

        exp_ret_list = list(
            np.arange(self.__min_expected_return, self.__max_expected_return,
                      step))
        std_list = [
            investment_universe.feasible_portfolios_equation_with_r_f(mu)
            for mu in exp_ret_list]

        legend_label = self.__complete_label("Feasible portfolios with r_f",
                                             label)
        if len(investment_universe.assets) < 2:
            self.__ax.plot(std_list, exp_ret_list, linewidth=size,
                           alpha=self.__alpha, label=legend_label, color=color,
                           zorder=zorder)
        else:
            max_std = (1 + self.FILL_BETWEEN_X_EXCESS) \
                      * self.__max_standard_deviation
            self.__ax.fill_betweenx(exp_ret_list, std_list, max_std,
                                    linewidth=size,
                                    alpha=self.__alpha,
                                    label=legend_label,
                                    color=color, zorder=zorder)

    def __plot_efficient_frontier(self, investment_universe, label, size,
                                  zorder):
        color_label = label if label is not None else "1"
        color = self.__colors[color_label]["efficient"]

        efficient_portfolios_mu_list = list(map(lambda x: x.expected_return,
                                                investment_universe.efficient_frontier))
        efficient_portfolios_std_list = list(
            map(lambda x: x.standard_deviation,
                investment_universe.efficient_frontier))
        legend_label = self.__complete_label("Efficient frontier", label)

        self.__ax.plot(efficient_portfolios_std_list,
                       efficient_portfolios_mu_list,
                       linewidth=size, alpha=self.__alpha, label=legend_label,
                       color=color, zorder=zorder)

    def __plot_efficient_frontier_equation(self, investment_universe, label,
                                           size, zorder):
        color_label = label if label is not None else "1"
        color = self.__colors[color_label]["efficient"]

        efficient_frontier_equation = \
            investment_universe.efficient_frontier_equation["equation"]
        mvp_mu_sigma = investment_universe.efficient_frontier_equation["mvp"]
        mvp_mu = mvp_mu_sigma[0]

        nb_points = self.LINE_NB_POINTS
        step = (self.__max_expected_return - mvp_mu) / nb_points

        exp_ret_list = list(
            np.arange(mvp_mu, self.__max_expected_return, step))
        std_list = [efficient_frontier_equation(mu) for mu in exp_ret_list]

        legend_label = self.__complete_label("Efficient frontier", label)
        self.__ax.plot(std_list, exp_ret_list, linewidth=size,
                       alpha=self.__alpha, label=legend_label, color=color,
                       zorder=zorder)

    def __plot_cal(self, investment_universe, label, size, zorder):
        color_label = label if label is not None else "1"
        color = self.__colors[color_label]["cal"]

        cal_portfolios_mu_list = list(map(lambda x: x.expected_return,
                                          investment_universe.cal))
        cal_portfolios_std_list = list(map(lambda x: x.standard_deviation,
                                           investment_universe.cal))
        legend_label = self.__complete_label("CAL", label)
        self.__ax.plot(cal_portfolios_std_list, cal_portfolios_mu_list,
                       linewidth=size, alpha=self.__alpha, label=legend_label,
                       color=color, zorder=zorder)

    def __plot_cal_equation(self, investment_universe, label, size, zorder):
        color_label = label if label is not None else "1"
        color = self.__colors[color_label]["cal"]

        nb_points = self.LINE_NB_POINTS
        step = (
                       self.__max_expected_return - self.__min_expected_return) / nb_points

        exp_ret_list = list(
            np.arange(investment_universe.r_f, self.__max_expected_return,
                      step))
        std_list = [investment_universe.cal_equation(mu)
                    for mu in exp_ret_list]

        legend_label = self.__complete_label("CAL", label)
        self.__ax.plot(std_list, exp_ret_list, linewidth=size,
                       alpha=self.__alpha, label=legend_label, color=color,
                       zorder=zorder)

    def __plot_assets(self, label, size, zorder):
        color_label = label if label is not None else "1"
        color_iter = iter(self.__colors[color_label]["assets"])

        for asset, (inv_univ_index, asset_index) \
                in self.__assets_inv_univ_dict.items():
            inv_univ = self.__investment_universes[inv_univ_index]
            color = next(color_iter)

            weights = np.zeros(len(inv_univ.assets)) if inv_univ.r_f is None \
                else np.zeros(len(inv_univ.assets) + 1)
            weights[asset_index] = 1
            weights_str = \
                str([("{:." + str(self.__nb_decimal_places) + "f}").format(
                    round(weight, self.__nb_decimal_places)) for weight
                    in weights]).replace("'", "")

            legend_label = inv_univ.assets[asset_index] + " - " + weights_str

            self.__ax.scatter(inv_univ.std[asset_index],
                              inv_univ.mu[asset_index], s=size,
                              label=legend_label,
                              color=color, zorder=zorder)

    def __plot_mvp(self, investment_universe, label, size, zorder):
        color_label = label if label is not None else "1"
        color = self.__colors[color_label]["mvp"]

        weights = str([("{:." + str(self.__nb_decimal_places) + "f}").format(
            round(weight, self.__nb_decimal_places)) for weight
            in investment_universe.mvp.weights]).replace("'", "")

        legend_label = self.__complete_label("MVP - " + weights, label)
        self.__ax.scatter(investment_universe.mvp.standard_deviation,
                          investment_universe.mvp.expected_return, s=size,
                          label=legend_label, color=color, zorder=zorder)

    def __plot_r_f(self, investment_universe, label, size, zorder):
        color_label = label if label is not None else "1"
        color = self.__colors[color_label]["r_f"]

        weights = np.zeros(len(investment_universe.assets) + 1)
        weights[-1] = 1
        weights_str = \
            str([("{:." + str(self.__nb_decimal_places) + "f}").format(
                round(weight, self.__nb_decimal_places)) for weight
                in weights]).replace("'", "")

        legend_label = "r_f - " + weights_str

        self.__ax.scatter(0, investment_universe.r_f, s=size,
                          label=legend_label, color=color, zorder=zorder)

    def __plot_tangency_portfolio(self, investment_universe, label, size,
                                  zorder):
        color_label = label if label is not None else "1"
        color = self.__colors[color_label]["tangency"]

        weights = str([("{:." + str(self.__nb_decimal_places) + "f}").format(
            round(weight, self.__nb_decimal_places)) for weight
            in investment_universe.tangency_portfolio.weights]).replace(
            "'", "")

        legend_label = self.__complete_label("Tangency portfolio - " + weights,
                                             label)
        self.__ax.scatter(
            investment_universe.tangency_portfolio.standard_deviation,
            investment_universe.tangency_portfolio.expected_return, s=size,
            label=legend_label, color=color, zorder=zorder)

    def __plot_other_portfolios(self, investment_universe, label, size,
                                zorder):
        color_label = label if label is not None else "1"
        color_iter = iter(self.__colors[color_label]["others"])

        for portfolio_weights, portfolio_with_name_pair \
                in investment_universe.other_portfolios.items():
            portfolio = portfolio_with_name_pair[0]
            portfolio_name = portfolio_with_name_pair[1]

            weights = str([("{:." + str(self.__nb_decimal_places) + "f}")
                          .format(round(weight, self.__nb_decimal_places))
                           for weight in portfolio_weights]).replace("'", "")

            if portfolio_name is not None:
                name_weights = portfolio_name + " - " + str(weights) \
                    if self.__weights_visible else portfolio_name
            else:
                name_weights = str(weights)

            legend_label = self.__complete_label(name_weights, label)

            color = next(color_iter)

            self.__ax.scatter(portfolio.standard_deviation,
                              portfolio.expected_return, s=size,
                              label=legend_label, color=color, zorder=zorder)

    def __plot_market_portfolio(self, investment_universe, label, size,
                                zorder):
        color_label = label if label is not None else "1"
        color = self.__colors[color_label]["market"]

        weights = str([("{:." + str(self.__nb_decimal_places) + "f}").format(
            round(weight, self.__nb_decimal_places)) for weight
            in investment_universe.market_portfolio.weights]).replace(
            "'", "")

        legend_label = self.__complete_label("Market portfolio - " + weights,
                                             label)
        self.__ax.scatter(
            investment_universe.market_portfolio.standard_deviation,
            investment_universe.market_portfolio.expected_return, s=size,
            label=legend_label, color=color, zorder=zorder)

    def __plot_investors_portfolio(self, investment_universe, investors,
                                   indifference_curves, label, size, zorder):

        for investor_name in investors:
            if investor_name in investment_universe.investors:
                investor = investment_universe.investors[investor_name]
                weights = str(
                    [("{:." + str(self.__nb_decimal_places) + "f}").format(
                        round(weight, self.__nb_decimal_places)) for weight
                        in investor.portfolio.weights]).replace(
                    "'", "")

                legend_label = self.__complete_label(
                    investor_name + " - " + weights,
                    label)

                color = self.__get_investor_color(investor_name, label)
                self.__ax.scatter(
                    investor.portfolio.standard_deviation,
                    investor.portfolio.expected_return,
                    s=size, label=legend_label, color=color, zorder=zorder)

    def __plot_indifference_curves(self, investment_universe, investors,
                                   indifference_curves, label, size, zorder,
                                   default_nb_curves):

        # Determine the number of indifference curves
        if indifference_curves is True:
            nb_curves = default_nb_curves
        elif type(indifference_curves) is int:
            nb_curves = indifference_curves
        else:
            nb_curves = None

        # If indifference_curves is not a list, then there is only one
        # indifference curve
        if type(indifference_curves) is not list:
            indifference_curves = [indifference_curves]

        if investors is None:
            investors = investment_universe.investors

        for investor_name in investors:
            if investor_name in investment_universe.investors:
                investor = investment_universe.investors[investor_name]

                if nb_curves is not None:
                    start_utility = (self.max_expected_return
                                     - self.min_expected_return) / 2
                    utility_list = self.__get_default_utility_list(
                        start_utility, nb_curves)
                else:
                    utility_list = self.__get_indifference_curves_utility_list(
                        indifference_curves, investment_universe, investor)

                color = self.__get_investor_color(investor_name, label)
                legend_label = self.__complete_label(investor_name, label)
                utility_curve_nb = 1
                for utility in utility_list:
                    std_array, mu_array = \
                        investor.calculate_indifference_curve(
                            utility, min_std=self.__min_standard_deviation,
                            max_std=self.__max_standard_deviation)

                    if utility_curve_nb == 1:
                        self.__ax.plot(std_array, mu_array, linewidth=size,
                                       alpha=self.__alpha, color=color,
                                       zorder=zorder, label=legend_label)
                    else:
                        self.__ax.plot(std_array, mu_array, linewidth=size,
                                       alpha=self.__alpha, color=color,
                                       zorder=zorder)
                    utility_curve_nb += 1

    def __get_default_utility_list(self, portfolio_utility, nb_curves):

        min_utility = 2 * self.__min_expected_return \
                      - self.__max_expected_return
        max_utility = self.__max_expected_return
        utility_step = (self.__max_expected_return
                        - self.__min_expected_return) / nb_curves

        utility_list = []

        utility = portfolio_utility
        utility_list.append(utility)
        while utility > min_utility:
            utility -= utility_step
            utility_list.append(utility)

        while utility < max_utility:
            utility += utility_step
            utility_list.append(utility)

        return utility_list

    def __get_indifference_curves_utility_list(self, indifference_curves,
                                               investment_universe, investor):

        utility_list = []

        for ind_curve in indifference_curves:
            utility = None
            if type(ind_curve) is float or type(ind_curve) is int:
                utility = ind_curve
            elif type(ind_curve) is list:
                ind_curve_ptf = pyvest.general.Portfolio(
                    ind_curve, investment_universe.mu,
                    investment_universe.cov,
                    r_f=investment_universe.r_f,
                    assets=investment_universe.assets)
                utility = investor.calculate_portfolio_utility(ind_curve_ptf)
            elif type(ind_curve) is str:
                ind_curve_ptf = self.__get_portfolio_from_string(
                    ind_curve, investment_universe, investor)
                if ind_curve_ptf is not None:
                    utility = investor.calculate_portfolio_utility(
                        ind_curve_ptf)

            if utility is not None:
                utility_list.append(utility)

        return utility_list

    def __get_portfolio_from_string(self, ptf_string, investment_universe,
                                    investor):

        if ptf_string in investment_universe.assets:
            asset_index = investment_universe.assets.index(ptf_string)
            ptf_weights = np.zeros(len(investment_universe.assets))
            ptf_weights[asset_index] = 1
            ptf = pyvest.general.Portfolio(
                ptf_weights, investment_universe.mu,
                investment_universe.cov,
                r_f=investment_universe.r_f,
                assets=investment_universe.assets)
        elif ptf_string == "mvp" and investment_universe.mvp is not None:
            ptf = investment_universe.mvp
        elif ptf_string == "tangency" \
                and investment_universe.tangency_portfolio is not None:
            ptf = investment_universe.tangency_portfolio
        elif ptf_string == "optimal" \
                and investor.optimal_portfolio is not None:
            ptf = investor.optimal_portfolio
        elif ptf_string == investor.name \
                and investor.optimal_portfolio is not None:
            ptf = investor.optimal_portfolio
        else:
            ptf = investment_universe.get_portfolio_by_name(ptf_string)

        return ptf

    def __plot_optimal_portfolios(self, investment_universe, investors,
                                  optimal_portfolios, label, size, zorder):

        if investors is None:
            investors = investment_universe.investors

        if optimal_portfolios is True:
            optimal_ptf_investors = investors
        elif optimal_portfolios is False:
            optimal_ptf_investors = []
        else:
            optimal_portfolios = [optimal_portfolios] \
                if type(optimal_portfolios) is not list else optimal_portfolios
            optimal_ptf_investors = list(set(investors).intersection(
                set(optimal_portfolios)))

        for investor_name in optimal_ptf_investors:
            if investor_name in investment_universe.investors:
                investor = investment_universe.investors[investor_name]
                if investor.optimal_portfolio is not None:
                    weights = str(
                        [("{:." + str(self.__nb_decimal_places) + "f}").format(
                            round(weight, self.__nb_decimal_places)) for weight
                            in investor.optimal_portfolio.weights]).replace(
                        "'", "")

                    opt_label_str = "Optimal - " + investor_name + " - " + weights
                    legend_label = self.__complete_label(opt_label_str, label)

                    color = self.__get_investor_color(investor_name, label)
                    self.__ax.scatter(
                        investor.optimal_portfolio.standard_deviation,
                        investor.optimal_portfolio.expected_return,
                        s=size, label=legend_label, color=color, zorder=zorder)

    def __check_bool(self, value, variable_name):
        if type(value) is not bool:
            raise TypeError("'{}' must be a boolean!".format(variable_name))

    def __complete_label(self, initial_legend_label, additional_label):
        completed_legend_label = initial_legend_label
        if additional_label is not None:
            completed_legend_label += " - " + additional_label

        return completed_legend_label

    def reset_colors(self):

        tab20_cmap = matplotlib.cm.tab20
        tab20b_cmap = matplotlib.cm.tab20b
        tab20c_cmap = matplotlib.cm.tab20c
        dark2_cmap = matplotlib.cm.Dark2
        accent_cmap = matplotlib.cm.Accent

        colors1 = {
            'feasible': tab20_cmap(0),
            'feasible_with_r_f': tab20_cmap(3),
            'mvp': tab20_cmap(6),
            'efficient': 'black',
            'tangency': tab20_cmap(4),
            'r_f': tab20c_cmap(16),
            'cal': tab20_cmap(2),
            'assets': [tab20b_cmap(i) for i in range(0, 20, 4)],
            'others': [tab20b_cmap(i) for i in range(1, 20, 4)] + [
                dark2_cmap(i) for i in range(0, 8)],
            'market': 'yellow',
            'investors': [dark2_cmap(i) for i in range(0, 8, 1)]
        }

        colors2 = {
            'feasible': tab20_cmap(1),
            'feasible_with_r_f': tab20_cmap(13),
            'mvp': tab20_cmap(7),
            'efficient': 'black',
            'tangency': tab20_cmap(5),
            'r_f': tab20c_cmap(16),
            'cal': tab20_cmap(12),
            'assets': [tab20b_cmap(i) for i in range(2, 20, 4)],
            'others': [tab20b_cmap(i) for i in range(3, 20, 4)] + [
                accent_cmap(i) for i in range(0, 8)],
            'market': 'lightskyblue',
            'investors': [dark2_cmap(i) for i in range(0, 8)]
        }

        # TODO: Modify colors3 and colors4 so that the colors are all
        #  different from colors1 and colors2
        colors3 = {
            'feasible': tab20_cmap(2),
            'feasible_with_r_f': tab20_cmap(15),
            'mvp': tab20_cmap(8),
            'efficient': 'black',
            'tangency': tab20_cmap(9),
            'r_f': tab20c_cmap(16),
            'cal': tab20_cmap(14),
            'assets': [tab20b_cmap(i) for i in range(3, 20, 4)],
            'others': [dark2_cmap(i) for i in range(0, 8)] + [
                accent_cmap(i) for i in range(0, 8)],
            'market': 'hotpink',
            'investors': [accent_cmap(i) for i in range(0, 8)]
        }

        colors4 = {
            'feasible': accent_cmap(0),
            'feasible_with_r_f': tab20_cmap(17),
            'mvp': accent_cmap(1),
            'efficient': 'black',
            'tangency': accent_cmap(2),
            'r_f': tab20c_cmap(16),
            'cal': tab20_cmap(16),
            'assets': [tab20b_cmap(i) for i in range(1, 20, 4)],
            'others': [dark2_cmap(i) for i in range(4, 8)] + [
                accent_cmap(i) for i in range(4, 8)],
            'market': 'palegreen',
            'investors': [accent_cmap(i) for i in range(4, 8)]
        }

        if len(self.__labels) > 0:
            self.__colors = {
                self.__labels[0]: colors1,
                self.__labels[1]: colors2,
                self.__labels[2]: colors3,
                self.__labels[3]: colors4
            }
        else:
            self.__colors = {
                "1": colors1,
                "2": colors2,
                "3": colors3,
                "4": colors4
            }

    def __set_default_visual_elements_properties(self):

        self.__visual_elements_properties = {
            "assets": {
                "priority": 10,
                "size": 200
            }
        }

        for inv_univ_index in range(0, self.MAX_NB_INV_UNIV):
            vis_elem_properties = {
                "r_f": {
                    "priority": 5 - inv_univ_index,
                    "size": 200
                },
                "investors": {
                    "priority": 10 - inv_univ_index,
                    "size": 200
                },
                "others": {
                    "priority": 15 - inv_univ_index,
                    "size": 200
                },
                "market_portfolio": {
                    "priority": 30 - inv_univ_index,
                    "size": 200
                },
                "tangency_portfolio": {
                    "priority": 40 - inv_univ_index,
                    "size": 200
                },
                "mvp": {
                    "priority": 50 - inv_univ_index,
                    "size": 200
                },
                "cal": {
                    "priority": 60 - inv_univ_index,
                    "size": 7
                },
                "efficient_portfolios": {
                    "priority": 70 - inv_univ_index,
                    "size": 7
                },
                "feasible_portfolios": {
                    "priority": 80 - inv_univ_index,
                    "size": 0
                },
                "feasible_portfolios_with_r_f": {
                    "priority": 90 - inv_univ_index,
                    "size": 0
                },
                "indifference_curves": {
                    "priority": 20 - inv_univ_index,
                    "size": 2,
                    "nb_curves": 7
                },
                "optimal_portfolios": {
                    "priority": 15 - inv_univ_index,
                    "size": 200
                }
            }
            self.__visual_elements_properties[
                inv_univ_index] = vis_elem_properties

    def __generate_visual_elements_list(self, investors, indifference_curves,
                                        optimal_portfolios):

        self.__visual_elements_list = []

        assets_properties = self.__visual_elements_properties["assets"]
        if self.__assets_visible:
            assets_label = self.__labels[0] if len(self.__labels) > 0 else None
            self.__visual_elements_list.append(
                self.VisualElement(self.__plot_assets,
                                   assets_properties["priority"],
                                   assets_properties["size"],
                                   label=assets_label))

        inv_univ_index = 0
        labels_iter = iter(self.__labels)
        for inv_univ in self.__investment_universes:
            label = next(labels_iter, None)
            properties = self.__visual_elements_properties[inv_univ_index]
            if inv_univ.other_portfolios \
                    and self.__other_portfolios_visible:
                self.__visual_elements_list.append(
                    self.VisualElement(self.__plot_other_portfolios,
                                       properties["others"]["priority"],
                                       properties["others"]["size"], inv_univ,
                                       label))
            if inv_univ.r_f and self.__r_f_visible:
                self.__visual_elements_list.append(
                    self.VisualElement(self.__plot_r_f,
                                       properties["r_f"]["priority"],
                                       properties["r_f"]["size"], inv_univ,
                                       label))
            if inv_univ.tangency_portfolio \
                    and self.__tangency_portfolio_visible:
                self.__visual_elements_list.append(
                    self.VisualElement(self.__plot_tangency_portfolio,
                                       properties["tangency_portfolio"][
                                           "priority"],
                                       properties["tangency_portfolio"][
                                           "size"], inv_univ, label))
            if inv_univ.mvp \
                    and self.__mvp_visible:
                self.__visual_elements_list.append(
                    self.VisualElement(self.__plot_mvp,
                                       properties["mvp"]["priority"],
                                       properties["mvp"]["size"], inv_univ,
                                       label))
            if inv_univ.cal \
                    and self.__cal_visible:
                self.__visual_elements_list.append(
                    self.VisualElement(self.__plot_cal,
                                       properties["cal"]["priority"],
                                       properties["cal"]["size"], inv_univ,
                                       label))
            if inv_univ.cal_equation \
                    and self.__cal_visible:
                self.__visual_elements_list.append(
                    self.VisualElement(self.__plot_cal_equation,
                                       properties["cal"]["priority"],
                                       properties["cal"]["size"], inv_univ,
                                       label))
            if inv_univ.efficient_frontier \
                    and self.__efficient_frontier_visible:
                self.__visual_elements_list.append(
                    self.VisualElement(self.__plot_efficient_frontier,
                                       properties["efficient_portfolios"][
                                           "priority"],
                                       properties["efficient_portfolios"][
                                           "size"], inv_univ, label))
            if inv_univ.efficient_frontier_equation \
                    and self.__efficient_frontier_visible:
                self.__visual_elements_list.append(
                    self.VisualElement(self.__plot_efficient_frontier_equation,
                                       properties["efficient_portfolios"][
                                           "priority"],
                                       properties["efficient_portfolios"][
                                           "size"], inv_univ, label))
            if inv_univ.feasible_portfolios \
                    and self.__feasible_portfolios_visible:
                self.__visual_elements_list.append(
                    self.VisualElement(self.__plot_feasible_portfolios,
                                       properties["feasible_portfolios"][
                                           "priority"],
                                       properties["feasible_portfolios"][
                                           "size"], inv_univ, label))
            if inv_univ.feasible_portfolios_with_r_f \
                    and self.__feasible_portfolios_with_r_f_visible:
                self.__visual_elements_list.append(
                    self.VisualElement(
                        self.__plot_feasible_portfolios_with_r_f,
                        properties["feasible_portfolios_with_r_f"]["priority"],
                        properties["feasible_portfolios_with_r_f"]["size"],
                        inv_univ, label))
            if inv_univ.feasible_portfolios_surface \
                    and self.__feasible_portfolios_visible:
                self.__visual_elements_list.append(
                    self.VisualElement(self.__plot_feasible_portfolios_surface,
                                       properties["feasible_portfolios"][
                                           "priority"],
                                       properties["feasible_portfolios"][
                                           "size"], inv_univ, label))
            if inv_univ.feasible_portfolios_surface_with_r_f \
                    and self.__feasible_portfolios_visible:
                self.__visual_elements_list.append(
                    self.VisualElement(
                        self.__plot_feasible_portfolios_surface_with_r_f,
                        properties["feasible_portfolios_with_r_f"][
                            "priority"],
                        properties["feasible_portfolios_with_r_f"][
                            "size"], inv_univ, label))
            if inv_univ.feasible_portfolios_equation \
                    and self.__feasible_portfolios_visible:
                self.__visual_elements_list.append(
                    self.VisualElement(
                        self.__plot_feasible_portfolios_equation,
                        properties["feasible_portfolios"][
                            "priority"],
                        properties["feasible_portfolios"][
                            "size"], inv_univ, label))
            if inv_univ.feasible_portfolios_equation_with_r_f \
                    and self.__feasible_portfolios_visible:
                self.__visual_elements_list.append(
                    self.VisualElement(
                        self.__plot_feasible_portfolios_equation_with_r_f,
                        properties["feasible_portfolios_with_r_f"][
                            "priority"],
                        properties["feasible_portfolios_with_r_f"][
                            "size"], inv_univ, label))
            if inv_univ.market_portfolio \
                    and self.__market_portfolio_visible:
                self.__visual_elements_list.append(
                    self.VisualElement(self.__plot_market_portfolio,
                                       properties["market_portfolio"][
                                           "priority"],
                                       properties["market_portfolio"][
                                           "size"], inv_univ, label))

            if indifference_curves is not None \
                    and indifference_curves is not False:
                self.__visual_elements_list.append(
                    self.VisualElement(
                        self.__plot_indifference_curves,
                        properties["indifference_curves"]["priority"],
                        properties["indifference_curves"]["size"],
                        inv_univ, label, investors, indifference_curves,
                        properties["indifference_curves"]["nb_curves"]))

            if optimal_portfolios is not None \
                    and optimal_portfolios is not False:
                self.__visual_elements_list.append(
                    self.VisualElement(
                        self.__plot_optimal_portfolios,
                        properties["optimal_portfolios"]["priority"],
                        properties["optimal_portfolios"]["size"],
                        inv_univ, label, investors, optimal_portfolios))

            inv_univ_index += 1

    def __get_investor_color(self, investor_name, label):

        if investor_name in self.__investors_colors:
            investor_color = self.__investors_colors[investor_name]
        else:
            color_label = label if label is not None else "1"
            color_iter = iter(self.__colors[color_label]["investors"])

            investor_color = next(color_iter)
            while investor_color in self.__investors_colors.values():
                investor_color = next(color_iter)

            self.__investors_colors[investor_name] = investor_color

        return investor_color



import matplotlib
import matplotlib.pyplot as plt

import pyvest.investment as investment_module


class InvestmentVisualizer:
    MAX_NB_INVESTMENTS = 4

    class VisualElement:
        def __init__(self, plot_function, priority, data=None, style=None):
            self.__plot_function = plot_function
            self.__data = data
            self.__style = style
            self.__priority = priority

            self.__zorder = 1 / self.__priority

        def plot(self):
            self.__plot_function(self.__data, self.__style,
                                 self.__zorder)

        def __lt__(self, other):
            return self.__priority < other.__priority

    def __init__(self, investment, labels=None):

        self.__assign_investments(investment)
        self.__assign_labels(labels)
        self.__set_default_colors()
        self.__set_default_visual_elements_properties()

        self.__ax = None
        self.__fig = None

    @property
    def investments(self):
        return self.__investments

    @investments.setter
    def investments(self, value):
        self.__assign_investments(value)
        self.__assign_labels(None)

    @property
    def labels(self):
        return self.__labels

    @labels.setter
    def labels(self, value):
        self.__assign_labels(value)

    @property
    def fig(self):
        return self.__fig

    @property
    def ax(self):
        return self.__ax

    def plot_profit(self, start_date=None, end_date=None, figsize=(16, 9),
                    legend='best', show_transactions=False):
        self.__prepare_profit_plot(figsize)

        visual_elements = self.__generate_profits_visual_elements(
            start_date, end_date)

        transactions_visual_elements = []
        if show_transactions:
            transactions_visual_elements = \
                self.__generate_transactions_visual_elements(start_date,
                                                             end_date)

        sorted_visual_elements = \
            sorted(visual_elements + transactions_visual_elements,
                   reverse=True)
        for vis_elem in sorted_visual_elements:
            vis_elem.plot()

        if type(legend) is str:
            self.__ax.legend(fontsize=15, loc=legend)

    def plot_return(self, start_date=None, end_date=None, figsize=(16, 9),
                    legend='best', show_transactions=False):
        self.__prepare_return_plot(figsize)

        visual_elements = self.__generate_returns_visual_elements(
            start_date, end_date)

        transactions_visual_elements = []
        if show_transactions:
            transactions_visual_elements = \
                self.__generate_transactions_visual_elements(start_date,
                                                             end_date)

        sorted_visual_elements = \
            sorted(visual_elements + transactions_visual_elements,
                   reverse=True)
        for vis_elem in sorted_visual_elements:
            vis_elem.plot()

        if type(legend) is str:
            self.__ax.legend(fontsize=15, loc=legend)

    def plot_weight(self, start_date=None, end_date=None, figsize=(16, 9),
                    legend='best', show_transactions=False):
        self.__prepare_weight_plot(figsize)

        visual_elements = self.__generate_weight_visual_elements(
            start_date, end_date)

        transactions_visual_elements = []
        if show_transactions:
            transactions_visual_elements = \
                self.__generate_transactions_visual_elements(start_date,
                                                             end_date)

        sorted_visual_elements = \
            sorted(visual_elements + transactions_visual_elements,
                   reverse=True)
        for vis_elem in sorted_visual_elements:
            vis_elem.plot()

        if type(legend) is str:
            self.__ax.legend(fontsize=15, loc=legend)

    def plot_value(self, start_date=None, end_date=None, figsize=(16, 9),
                    legend='best', show_transactions=False):
        self.__prepare_value_plot(figsize)

        visual_elements = self.__generate_value_visual_elements(
            start_date, end_date)

        transactions_visual_elements = []
        if show_transactions:
            transactions_visual_elements = \
                self.__generate_transactions_visual_elements(start_date,
                                                             end_date)

        sorted_visual_elements = \
            sorted(visual_elements + transactions_visual_elements,
                   reverse=True)
        for vis_elem in sorted_visual_elements:
            vis_elem.plot()

        if type(legend) is str:
            self.__ax.legend(fontsize=15, loc=legend)

    def __prepare_profit_plot(self, figsize):
        self.__fig, self.__ax = plt.subplots(figsize=figsize)
        self.__ax.grid()

        self.__ax.set_title("Profit", fontsize=35)
        self.__ax.set_ylabel("Value ($)", fontsize=30)
        self.__ax.set_xlabel("Date", fontsize=30)
        self.__ax.tick_params(axis='both', labelsize=25)

    def __prepare_return_plot(self, figsize):
        self.__fig, self.__ax = plt.subplots(figsize=figsize)
        self.__ax.grid()

        self.__ax.set_title("Return", fontsize=35)
        self.__ax.set_ylabel("Percentage (%)", fontsize=30)
        self.__ax.set_xlabel("Date", fontsize=30)
        self.__ax.tick_params(axis='both', labelsize=25)

    def __prepare_weight_plot(self, figsize):
        self.__fig, self.__ax = plt.subplots(figsize=figsize)
        self.__ax.grid()

        self.__ax.set_title("Weight by asset", fontsize=35)
        self.__ax.set_ylabel("Weight", fontsize=30)
        self.__ax.set_xlabel("Date", fontsize=30)
        self.__ax.tick_params(axis='both', labelsize=25)

    def __prepare_value_plot(self, figsize):
        self.__fig, self.__ax = plt.subplots(figsize=figsize)
        self.__ax.grid()

        self.__ax.set_title("Value by asset", fontsize=35)
        self.__ax.set_ylabel("Value ($)", fontsize=30)
        self.__ax.set_xlabel("Date", fontsize=30)
        self.__ax.tick_params(axis='both', labelsize=25)

    def __assign_investments(self, investments):
        if isinstance(investments, investment_module.Investment):
            self.__investments = [investments]
        else:
            self.__investments = investments

    def __assign_labels(self, labels):
        generic_labels = ["1", "2", "3", "4"]
        if labels is None and len(self.__investments) > 1:
            self.__labels = generic_labels
        elif labels is None:
            self.__labels = []
        else:
            self.__labels = labels + generic_labels[len(labels):]

    def __set_default_colors(self):

        tab10_cmap = matplotlib.cm.tab10
        tab20b_cmap = matplotlib.cm.tab20b
        set2_cmap = matplotlib.cm.Set2
        dark2_cmap = matplotlib.cm.Dark2

        colors1 = {
            'unrealized_capital_gain': tab10_cmap(0),
            'realized_capital_gain': tab10_cmap(1),
            'dividends': tab10_cmap(2),
            'total_profits': tab10_cmap(3),
            'buy_transactions': tab10_cmap(0),
            'sell_transactions': tab10_cmap(1),
            'dividend_transactions': tab10_cmap(2),
            'rebalance_transactions': tab10_cmap(4),
            'total_returns': dark2_cmap(0),
            'assets': [tab20b_cmap(i) for i in range(0, 20, 4)]
        }

        colors2 = {
            'unrealized_capital_gain': set2_cmap(4),
            'realized_capital_gain': set2_cmap(5),
            'dividends': set2_cmap(6),
            'total_profits': set2_cmap(7),
            'buy_transactions': set2_cmap(4),
            'sell_transactions': set2_cmap(5),
            'dividend_transactions': set2_cmap(6),
            'rebalance_transactions': set2_cmap(0),
            'total_returns': dark2_cmap(1),
            'weight': dark2_cmap(1),
            'assets': [tab20b_cmap(i) for i in range(2, 20, 4)]
        }

        colors3 = {
            'unrealized_capital_gain': dark2_cmap(0),
            'realized_capital_gain': dark2_cmap(1),
            'dividends': dark2_cmap(2),
            'total_profits': dark2_cmap(3),
            'assets': [tab20b_cmap(i) for i in range(3, 20, 4)]
        }

        colors4 = {
            'unrealized_capital_gain': dark2_cmap(4),
            'realized_capital_gain': dark2_cmap(5),
            'dividends': dark2_cmap(6),
            'total_profits': dark2_cmap(7),
            'assets': [tab20b_cmap(i) for i in range(1, 20, 4)]
        }

        if len(self.__labels) > 0:
            self.__colors = {
                self.__labels[0]: colors1,
                self.__labels[1]: colors2,
                self.__labels[2]: colors3,
                self.__labels[3]: colors4
            }
        else:
            self.__colors = {
                "1": colors1,
                "2": colors2,
                "3": colors3,
                "4": colors4
            }

    def __set_default_visual_elements_properties(self):

        self.__visual_elements_properties = {}

        for inv_univ_index in range(0, self.MAX_NB_INVESTMENTS):
            vis_elem_properties = {
                "unrealized_capital_gains": {
                    "priority": 30 - inv_univ_index,
                    "size": 3
                },
                "realized_capital_gains": {
                    "priority": 25 - inv_univ_index,
                    "size": 3
                },
                "dividends": {
                    "priority": 20 - inv_univ_index,
                    "size": 3
                },
                "total_profits": {
                    "priority": 15 - inv_univ_index,
                    "size": 3
                },
                "buy_transactions": {
                    "priority": 10 - inv_univ_index,
                    "size": 3
                },
                "sell_transactions": {
                    "priority": 10 - inv_univ_index,
                    "size": 3
                },
                "dividend_transactions": {
                    "priority": 10 - inv_univ_index,
                    "size": 3
                },
                "rebalance_transactions": {
                    "priority": 10 - inv_univ_index,
                    "size": 3
                },
                "total_returns": {
                    "priority": 15 - inv_univ_index,
                    "size": 3
                },
                "weight": {
                    "priority": 15 - inv_univ_index,
                    "size": 3
                },
                "value": {
                    "priority": 15 - inv_univ_index,
                    "size": 3
                }
            }
            self.__visual_elements_properties[
                inv_univ_index] = vis_elem_properties

    def __complete_label(self, initial_legend_label, additional_label):
        completed_legend_label = initial_legend_label
        if additional_label is not None:
            completed_legend_label += " - " + additional_label

        return completed_legend_label

    def __plot_unrealized_capital_gains(self, profits, style, zorder):

        label = style["label"]
        size = style["size"]

        dates = [profit.date for profit in profits]
        unrealized_capital_gains = [profit.unrealized_capital_gain
                                    for profit in profits]

        color_label = label if label is not None else "1"
        color = self.__colors[color_label]["unrealized_capital_gain"]
        legend_label = self.__complete_label("Unrealized Capital Gains", label)

        self.__ax.plot(dates, unrealized_capital_gains, label=legend_label,
                       color=color, linewidth=size, zorder=zorder)

    def __plot_realized_capital_gains(self, profits, style, zorder):

        label = style["label"]
        size = style["size"]

        dates = [profit.date for profit in profits]
        realized_capital_gains = [profit.realized_capital_gain
                                  for profit in profits]

        color_label = label if label is not None else "1"
        color = self.__colors[color_label]["realized_capital_gain"]
        legend_label = self.__complete_label("Realized Capital Gains", label)

        self.__ax.plot(dates, realized_capital_gains, label=legend_label,
                       color=color, linewidth=size, zorder=zorder)

    def __plot_dividends(self, data, style, zorder):

        label = style["label"]
        size = style["size"]

        profits = data["profits"]
        reinvest_dividends = data["reinvest_dividends"]

        dates = [profit.date for profit in profits]

        if reinvest_dividends:
            dividends = [profit.reinvested_dividends for profit in profits]
        else:
            dividends = [profit.non_reinvested_dividends for profit in profits]

        color_label = label if label is not None else "1"
        color = self.__colors[color_label]["dividends"]
        legend_label = self.__complete_label("Dividends", label)

        self.__ax.plot(dates, dividends, label=legend_label,
                       color=color, linewidth=size, zorder=zorder)

    def __plot_total_profits(self, profits, style, zorder):

        label = style["label"]
        size = style["size"]

        dates = [profit.date for profit in profits]

        total_profits = [profit.non_reinvested_dividends
                         + profit.reinvested_dividends
                         + profit.unrealized_capital_gain
                         + profit.realized_capital_gain
                         for profit in profits]

        color_label = label if label is not None else "1"
        color = self.__colors[color_label]["total_profits"]
        legend_label = self.__complete_label("Total Profits", label)

        self.__ax.plot(dates, total_profits, label=legend_label, color=color,
                       linewidth=size, zorder=zorder)

    def __plot_buy_transactions(self, data_dict, style, zorder):

        label = style["label"]
        size = style["size"]

        investment = data_dict["investment"]
        start_date = data_dict["start_date"]
        end_date = data_dict["end_date"]

        transactions = investment.get_transactions(transaction_type="BUY",
                                                   start_date=start_date,
                                                   end_date=end_date)

        color_label = label if label is not None else "1"
        color = self.__colors[color_label]["buy_transactions"]
        legend_label = self.__complete_label("Buy", label)

        nb_lines = 0
        for date, transactions_list in transactions.items():
            vline_label = legend_label if nb_lines == 0 else None
            self.__ax.axvline(x=date, label=vline_label, color=color,
                              linestyle="dotted",
                              linewidth=size, zorder=zorder)
            nb_lines += 1

    def __plot_sell_transactions(self, data_dict, style, zorder):

        label = style["label"]
        size = style["size"]

        investment = data_dict["investment"]
        start_date = data_dict["start_date"]
        end_date = data_dict["end_date"]

        transactions = investment.get_transactions(transaction_type="SELL",
                                                   start_date=start_date,
                                                   end_date=end_date)

        color_label = label if label is not None else "1"
        color = self.__colors[color_label]["sell_transactions"]
        legend_label = self.__complete_label("Sell", label)

        nb_lines = 0
        for date, transactions_list in transactions.items():
            vline_label = legend_label if nb_lines == 0 else None
            self.__ax.axvline(x=date, label=vline_label, color=color,
                              linestyle="dotted",
                              linewidth=size, zorder=zorder)
            nb_lines += 1

    def __plot_dividend_transactions(self, data_dict, style, zorder):

        label = style["label"]
        size = style["size"]

        investment = data_dict["investment"]
        start_date = data_dict["start_date"]
        end_date = data_dict["end_date"]

        transaction_type = "REINVESTED_DIVIDEND" \
            if investment.reinvest_dividends \
            else investment.reinvest_dividends
        transactions = investment.get_transactions(
            transaction_type=transaction_type, start_date=start_date,
            end_date=end_date)

        color_label = label if label is not None else "1"
        color = self.__colors[color_label]["dividend_transactions"]
        legend_label = self.__complete_label("Dividend", label)

        nb_lines = 0
        for date, transactions_list in transactions.items():
            vline_label = legend_label if nb_lines == 0 else None
            self.__ax.axvline(x=date, label=vline_label, color=color,
                              linestyle="dotted",
                              linewidth=size, zorder=zorder)
            nb_lines += 1

    def __plot_rebalance_transactions(self, data_dict, style, zorder):

        label = style["label"]
        size = style["size"]

        investment = data_dict["investment"]
        start_date = data_dict["start_date"]
        end_date = data_dict["end_date"]

        transactions = investment.get_transactions(
            transaction_type="REBALANCE", start_date=start_date,
            end_date=end_date)

        color_label = label if label is not None else "1"
        color = self.__colors[color_label]["rebalance_transactions"]
        legend_label = self.__complete_label("Rebalance", label)

        nb_lines = 0
        for date, transactions_list in transactions.items():
            vline_label = legend_label if nb_lines == 0 else None
            self.__ax.axvline(x=date, label=vline_label, color=color,
                              linestyle="dotted",
                              linewidth=size, zorder=zorder)
            nb_lines += 1

    def __plot_total_returns(self, returns, style, zorder):

        label = style["label"]
        size = style["size"]

        dates = [inv_return.date for inv_return in returns]
        total_returns = [inv_return.total_return for inv_return in returns]

        color_label = label if label is not None else "1"
        color = self.__colors[color_label]["total_returns"]
        legend_label = self.__complete_label("Total Return", label)

        self.__ax.plot(dates, total_returns, label=legend_label,
                       color=color, linewidth=size, zorder=zorder)

    def __plot_weight(self, dated_weights_list, style, zorder):

        label = style["label"]
        size = style["size"]
        color = style["color"]

        dates = [dated_weight.date for dated_weight in dated_weights_list]
        weights = [dated_weight.weight for dated_weight
                   in dated_weights_list]

        if len(dated_weights_list) > 0:
            ticker = dated_weights_list[0].ticker
        else:
            ticker = ""

        legend_label = self.__complete_label(ticker, label)

        self.__ax.plot(dates, weights, label=legend_label,
                       color=color, linewidth=size, zorder=zorder)

    def __plot_value(self, dated_values_list, style, zorder):

        label = style["label"]
        size = style["size"]
        color = style["color"]

        dates = [dated_value.date for dated_value in dated_values_list]
        values = [dated_value.value for dated_value
                   in dated_values_list]

        if len(dated_values_list) > 0:
            ticker = dated_values_list[0].ticker
        else:
            ticker = ""

        legend_label = self.__complete_label(ticker, label)

        self.__ax.plot(dates, values, label=legend_label,
                       color=color, linewidth=size, zorder=zorder)

    def __generate_profits_visual_elements(self, start_date, end_date):
        visual_elements = []

        investment_index = 0
        labels_iter = iter(self.__labels)
        for investment in self.__investments:
            profits = investment.calculate_profit_range(start_date=start_date,
                                                        end_date=end_date)

            label = next(labels_iter, None)
            properties = self.__visual_elements_properties[investment_index]

            visual_elements.append(
                self.VisualElement(self.__plot_unrealized_capital_gains,
                                   properties["unrealized_capital_gains"][
                                       "priority"],
                                   data=profits,
                                   style={
                                       "size": properties[
                                           "unrealized_capital_gains"]["size"],
                                       "label": label
                                   }))

            visual_elements.append(
                self.VisualElement(self.__plot_realized_capital_gains,
                                   properties["realized_capital_gains"][
                                       "priority"],
                                   data=profits,
                                   style={
                                       "size": properties[
                                           "realized_capital_gains"]["size"],
                                       "label": label
                                   }))

            visual_elements.append(
                self.VisualElement(self.__plot_dividends,
                                   properties["dividends"][
                                       "priority"],
                                   data={
                                       "profits": profits,
                                       "reinvest_dividends":
                                           investment.reinvest_dividends
                                   },
                                   style={
                                       "size": properties[
                                           "dividends"]["size"],
                                       "label": label
                                   }))

            visual_elements.append(
                self.VisualElement(self.__plot_total_profits,
                                   properties["total_profits"]["priority"],
                                   data=profits,
                                   style={
                                       "size": properties[
                                           "total_profits"]["size"],
                                       "label": label
                                   }))

            investment_index += 1

        return visual_elements

    def __generate_transactions_visual_elements(self, start_date, end_date):
        visual_elements = []

        investment_index = 0
        labels_iter = iter(self.__labels)
        for investment in self.__investments:
            data_dict = {"investment": investment,
                         "start_date": start_date,
                         "end_date": end_date}

            label = next(labels_iter, None)
            properties = self.__visual_elements_properties[investment_index]

            visual_elements.append(
                self.VisualElement(self.__plot_buy_transactions,
                                   properties["buy_transactions"][
                                       "priority"],
                                   data=data_dict,
                                   style={
                                       "size": properties[
                                           "buy_transactions"]["size"],
                                       "label": label
                                   }))

            visual_elements.append(
                self.VisualElement(self.__plot_sell_transactions,
                                   properties["sell_transactions"][
                                       "priority"],
                                   data=data_dict,
                                   style={
                                       "size": properties[
                                           "sell_transactions"]["size"],
                                       "label": label
                                   }))

            visual_elements.append(
                self.VisualElement(self.__plot_dividend_transactions,
                                   properties["dividend_transactions"][
                                       "priority"],
                                   data=data_dict,
                                   style={
                                       "size": properties[
                                           "dividend_transactions"]["size"],
                                       "label": label
                                   }))

            visual_elements.append(
                self.VisualElement(self.__plot_rebalance_transactions,
                                   properties["rebalance_transactions"][
                                       "priority"],
                                   data=data_dict,
                                   style={
                                       "size": properties[
                                           "rebalance_transactions"]["size"],
                                       "label": label
                                   }))

            investment_index += 1

        return visual_elements

    def __generate_returns_visual_elements(self, start_date, end_date):
        visual_elements = []

        investment_index = 0
        labels_iter = iter(self.__labels)
        for investment in self.__investments:
            returns = investment.calculate_return_range(start_date=start_date,
                                                        end_date=end_date)

            label = next(labels_iter, None)
            properties = self.__visual_elements_properties[investment_index]

            visual_elements.append(
                self.VisualElement(self.__plot_total_returns,
                                   properties["total_returns"][
                                       "priority"],
                                   data=returns,
                                   style={
                                       "size": properties[
                                           "total_returns"]["size"],
                                       "label": label
                                   }))

            investment_index += 1

        return visual_elements

    def __generate_weight_visual_elements(self, start_date, end_date):
        visual_elements = []

        investment_index = 0
        labels_iter = iter(self.__labels)
        for investment in self.__investments:
            dated_weight_by_ticker_list = investment.calculate_weight_range(
                start_date=start_date, end_date=end_date)

            label = next(labels_iter, None)
            properties = self.__visual_elements_properties[investment_index]

            dated_weights_list_by_ticker = {}
            for dated_weight_by_ticker in dated_weight_by_ticker_list:
                for ticker, dated_weight in dated_weight_by_ticker.items():
                    if ticker not in dated_weights_list_by_ticker:
                        dated_weights_list_by_ticker[ticker] = []
                    dated_weights_list_by_ticker[ticker].append(dated_weight)

            color_label = label if label is not None else "1"
            color_iter = iter(self.__colors[color_label]["assets"])
            for ticker, dated_weights_list \
                    in dated_weights_list_by_ticker.items():
                color = next(color_iter)
                visual_elements.append(
                    self.VisualElement(self.__plot_weight,
                                       properties["weight"][
                                           "priority"],
                                       data=dated_weights_list,
                                       style={
                                           "size": properties[
                                               "total_returns"]["size"],
                                           "label": label,
                                           "color": color
                                       }))

            investment_index += 1

        return visual_elements

    def __generate_value_visual_elements(self, start_date, end_date):
        visual_elements = []

        investment_index = 0
        labels_iter = iter(self.__labels)
        for investment in self.__investments:
            dated_value_by_ticker_list = investment.calculate_value_range(
                start_date=start_date, end_date=end_date)

            label = next(labels_iter, None)
            properties = self.__visual_elements_properties[investment_index]

            dated_values_list_by_ticker = {}
            for dated_value_by_ticker in dated_value_by_ticker_list:
                for ticker, dated_value in dated_value_by_ticker.items():
                    if ticker not in dated_values_list_by_ticker:
                        dated_values_list_by_ticker[ticker] = []
                    dated_values_list_by_ticker[ticker].append(dated_value)

            color_label = label if label is not None else "1"
            color_iter = iter(self.__colors[color_label]["assets"])
            for ticker, dated_values_list \
                    in dated_values_list_by_ticker.items():
                color = next(color_iter)
                visual_elements.append(
                    self.VisualElement(self.__plot_value,
                                       properties["value"][
                                           "priority"],
                                       data=dated_values_list,
                                       style={
                                           "size": properties[
                                               "total_returns"]["size"],
                                           "label": label,
                                           "color": color
                                       }))

            investment_index += 1

        return visual_elements



from pyvest.factor_model.factor_model import *
from pyvest.factor_model.factor_model_visualizer import *
from pyvest.factor_model.capm import *
from pyvest.factor_model.ff3f import *
from pyvest.factor_model.regression import *

__all__ = ['FactorModel', 'FactorModelVisualizer', 'CAPM', 'FF3F',
           'Regression']




class FactorsDataReader:
    def __init__(self):
        pass

    def read_factors(self, start_date, end_date):
        raise NotImplementedError("FactorsDataReader.read_factors has not "
                                  "been implemented.")



import numpy as np
import pandas as pd
import math

import yfinance as yf


###############################################################################
############################# GENERAL FUNCTIONS ###############################
###############################################################################

def calculate_portfolio_expected_return(weights, mu):
    # This function calculate the expected return of a portfolio
    # The first argument, "weights", is a column vector that contains the
    # fraction of wealth invested in the underlying assets
    # The second argument, "mu", is a column vector containing the expected
    # return of the underlying assets

    return float(np.dot(weights, mu))


def calculate_portfolio_standard_deviation(weights, cov):
    # This function calculate the standard deviation of a portfolio
    # The first argument, "weights", is a column vector that contains the
    # fraction of wealth invested in the underlying assets
    # The second argument, "cov", is a variance-covariance matrix

    return math.sqrt(np.dot(weights, np.dot(cov, weights)))


def calculate_portfolio_sharpe_ratio(x, mu, cov, r_f):
    r_p = calculate_portfolio_expected_return(x, mu)
    cov_p = calculate_portfolio_standard_deviation(x, cov)
    return (r_p - r_f) / cov_p


def standard_utility_function(x, mu, cov, gamma):
    mu_p = calculate_portfolio_expected_return(x, mu)
    cov_p = calculate_portfolio_standard_deviation(x, cov)
    return mu_p - gamma * cov_p ** 2


################################################################################################
######################## DOWNLOAD AND TRANSFORM DATA FROM YAHOO FINANCE ########################
################################################################################################


def get_monthly_returns_series(ticker, start_date, end_date):
    history_df = yf.Ticker(ticker).history(start=start_date, end=end_date,
                                           auto_adjust=False)
    monthly_history_df = history_df.resample('1M').last()
    monthly_returns_series = monthly_history_df['Adj Close'].pct_change() * 100

    return monthly_returns_series


def get_monthly_returns(tickers_list, start_date, end_date):
    monthly_returns_df = pd.DataFrame()
    for ticker in tickers_list:
        monthly_returns_series = get_monthly_returns_series(ticker, start_date,
                                                            end_date)
        monthly_returns_df[ticker] = monthly_returns_series

    return monthly_returns_df



import pyvest.investment as investment_module
from pyvest.investment.profit import Profit


class Transaction:
    def __init__(self, type, investment, asset, quantity=None,
                 asset_price=None, commission=None, date=None):
        self.__type = type

        self.__asset = asset

        if isinstance(asset, dict):
            self.__ticker = list(asset.keys())
        else:
            self.__ticker = self.__asset.ticker

        self.__investment = investment

        self._quantity = quantity
        self._asset_price = asset_price
        self._commission = commission
        self._date = date

    def __repr__(self):
        return self.__generate_output()

    def __str__(self):
        return self.__generate_output()

    @property
    def type(self):
        return self.__type

    @property
    def investment(self):
        return self.__investment

    @property
    def quantity(self):
        return self._quantity

    @property
    def asset_price(self):
        return self._asset_price

    @property
    def commission(self):
        return self._commission

    @property
    def date(self):
        return self._date

    @property
    def ticker(self):
        return self.__ticker

    def update_investment_state(self, investment_state_old):
        raise NotImplementedError("Transaction.update_investment_state has not"
                                  " been implemented.")

    def _create_profit(self, ticker, realized_capital_gain=None,
                       unrealized_capital_gain=None,
                       reinvested_dividends=None,
                       non_reinvested_dividends=None,
                       profit_to_copy=None):

        realized_capital_gain_by_ticker = \
            {ticker: realized_capital_gain} \
                if realized_capital_gain is not None else None

        unrealized_capital_gain_by_ticker = \
            {ticker: unrealized_capital_gain} \
                if unrealized_capital_gain is not None else None

        reinvested_dividends_by_ticker = \
            {ticker: reinvested_dividends} \
                if reinvested_dividends is not None else None

        non_reinvested_dividends_by_ticker = \
            {ticker: non_reinvested_dividends} \
                if non_reinvested_dividends is not None else None

        profit = Profit(self.__investment, date=self.date,
                        realized_capital_gain_by_ticker=
                        realized_capital_gain_by_ticker,
                        unrealized_capital_gain_by_ticker=
                        unrealized_capital_gain_by_ticker,
                        reinvested_dividends_by_ticker=
                        reinvested_dividends_by_ticker,
                        non_reinvested_dividends_by_ticker=
                        non_reinvested_dividends_by_ticker,
                        profit_to_copy=profit_to_copy)

        return profit

    def _create_investment_state(self, ticker, quantity=None, total_cost=None,
                                 average_cost=None, total_commission=None,
                                 asset_price=None, cash=None,
                                 total_invested=None, profit=None,
                                 investment_state_to_copy=None):
        quantity_by_ticker = {ticker: quantity} if quantity is not None \
            else None
        total_cost_by_ticker = {ticker: total_cost} \
            if total_cost is not None else None
        average_cost_by_ticker = {ticker: average_cost} \
            if average_cost is not None else None
        total_commission_by_ticker = {ticker: total_commission} \
            if total_commission is not None else None
        asset_price_by_ticker = {ticker: asset_price} \
            if asset_price is not None else None

        investment_state = investment_module.InvestmentState(
            self.__investment,
            quantity_by_ticker=quantity_by_ticker,
            total_cost_by_ticker=total_cost_by_ticker,
            average_cost_by_ticker=average_cost_by_ticker,
            total_commission_by_ticker=total_commission_by_ticker,
            asset_price_by_ticker=asset_price_by_ticker, cash=cash,
            total_invested=total_invested, profit=profit,
            investment_state_to_copy=investment_state_to_copy)

        return investment_state

    def __generate_output(self):
        asset_str = str(self.__asset)
        asset_str = asset_str.replace("\n", "\n  ")

        output = "Transaction type: " + self.__type + "\n" \
                 + "Asset: \n  " + asset_str + "\n" \
                 + "Quantity: " + str(self._quantity) + "\n" \
                 + "Price: " + str(self._asset_price) + "\n" \
                 + "Commission: " + str(self._commission)

        if self._date is not None:
            output += "\n" + "Date: " + str(self._date)

        return output


class BuyTransaction(Transaction):
    def __init__(self, investment, asset, quantity, asset_price, commission,
                 date=None):
        super().__init__("BUY", investment, asset, quantity, asset_price,
                         commission, date)

    def update_investment_state(self, investment_state_old):
        # Single asset transaction
        ticker = self.ticker

        total_cost_new = \
            investment_state_old.total_cost_by_ticker[ticker] \
            + self.quantity * self.asset_price
        total_commission_new = \
            investment_state_old.total_commission_by_ticker[ticker] \
            + self.commission
        quantity_new = investment_state_old.quantity_by_ticker[ticker] \
                       + self.quantity
        average_cost_new = total_cost_new / quantity_new

        # TODO: CHECK average_cost_new
        unrealized_capital_gain_new = \
            quantity_new * (self.asset_price - average_cost_new)

        cash_new = investment_state_old.cash - (
                self.quantity * self.asset_price + self.commission)
        additional_investment = self.quantity * self.asset_price \
                                + self.commission
        total_invested_new = \
            investment_state_old.total_invested + additional_investment

        profit = self._create_profit(
            ticker, unrealized_capital_gain=unrealized_capital_gain_new,
            profit_to_copy=investment_state_old.profit)

        investment_state = self._create_investment_state(
            ticker,
            quantity=quantity_new, total_cost=total_cost_new,
            average_cost=average_cost_new,
            total_commission=total_commission_new,
            asset_price=self.asset_price, cash=cash_new,
            total_invested=total_invested_new, profit=profit,
            investment_state_to_copy=investment_state_old)

        return investment_state


class SellTransaction(Transaction):
    def __init__(self, investment, asset, quantity, asset_price, commission,
                 date=None):
        super().__init__("SELL", investment, asset, quantity, asset_price,
                         commission, date)

    def update_investment_state(self, investment_state_old):
        # Single asset transaction
        ticker = self.ticker

        total_cost_new = \
            investment_state_old.total_cost_by_ticker[ticker] \
            - self.quantity \
            * investment_state_old.average_cost_by_ticker[ticker]
        total_commission_new = \
            investment_state_old.total_commission_by_ticker[ticker] \
            + self.commission
        quantity_new = investment_state_old.quantity_by_ticker[ticker] \
                       - self.quantity

        unrealized_capital_gain_new = \
            quantity_new * (self.asset_price
                            - investment_state_old.average_cost_by_ticker[
                                ticker])

        add_realized_capital_gain = \
            self.quantity * (self.asset_price -
                             investment_state_old.average_cost_by_ticker[
                                 ticker])
        realized_capital_gain_new = \
            investment_state_old.profit.realized_capital_gain_by_ticker[
                ticker] + add_realized_capital_gain

        cash_new = investment_state_old.cash + (
                self.quantity * self.asset_price - self.commission)

        profit = self._create_profit(
            ticker, realized_capital_gain=realized_capital_gain_new,
            unrealized_capital_gain=unrealized_capital_gain_new,
            profit_to_copy=investment_state_old.profit)

        investment_state = self._create_investment_state(
            ticker,
            quantity=quantity_new, total_cost=total_cost_new,
            total_commission=total_commission_new,
            asset_price=self.asset_price, cash=cash_new, profit=profit,
            investment_state_to_copy=investment_state_old)

        return investment_state


class DividendTransaction(Transaction):
    def __init__(self, investment, asset, asset_price, commission,
                 dividend_by_share, quantity=None, date=None):
        super().__init__("DIVIDEND", investment, asset, quantity, asset_price,
                         commission, date)

        self.__dividend_by_share = dividend_by_share

    def update_investment_state(self, investment_state_old):
        # Single asset transaction
        ticker = self.ticker

        self._quantity = investment_state_old.quantity_by_ticker[ticker]

        total_commission_new = \
            investment_state_old.total_commission_by_ticker[ticker] \
            + self.commission
        non_reinvested_dividends_new = \
            investment_state_old.profit.non_reinvested_dividends_by_ticker[
                ticker] + self._quantity * self.__dividend_by_share

        unrealized_capital_gain_new = \
            investment_state_old.quantity_by_ticker[ticker] \
            * (self.asset_price - investment_state_old.average_cost_by_ticker[
                ticker])

        cash_new = investment_state_old.cash + (
                self._quantity * self.__dividend_by_share - self.commission)

        profit = self._create_profit(
            ticker, unrealized_capital_gain=unrealized_capital_gain_new,
            non_reinvested_dividends=non_reinvested_dividends_new,
            profit_to_copy=investment_state_old.profit)

        investment_state = self._create_investment_state(
            ticker,
            total_commission=total_commission_new,
            asset_price=self.asset_price, cash=cash_new,
            profit=profit, investment_state_to_copy=investment_state_old)

        return investment_state


class ReinvestedDividendTransaction(Transaction):
    def __init__(self, investment, asset, asset_price, commission,
                 dividend_by_share, quantity=None, date=None):
        super().__init__("REINVESTED_DIVIDEND", investment, asset, quantity,
                         asset_price, commission, date)

        self.__dividend_by_share = dividend_by_share

    def update_investment_state(self, investment_state_old):
        # Single asset transaction
        ticker = self.ticker

        self._quantity = investment_state_old.quantity_by_ticker[ticker]

        total_commission_new = \
            investment_state_old.total_commission_by_ticker[ticker] \
            + self.commission
        additional_reinvested_dividends = \
            self._quantity * self.__dividend_by_share
        reinvested_dividends_new = \
            investment_state_old.profit.reinvested_dividends_by_ticker[
                ticker] + additional_reinvested_dividends

        total_cost_new = \
            investment_state_old.total_cost_by_ticker[ticker] \
            + additional_reinvested_dividends

        additional_quantity = \
            additional_reinvested_dividends / self.asset_price
        quantity_new = \
            investment_state_old.quantity_by_ticker[ticker] \
            + additional_quantity

        average_cost_new = total_cost_new / quantity_new if quantity_new > 0 \
            else 0

        unrealized_capital_gain_new = \
            quantity_new * (self.asset_price
                            - investment_state_old.average_cost_by_ticker[
                                ticker])

        additional_investment = \
            additional_reinvested_dividends + self.commission
        total_invested_new = \
            investment_state_old.total_invested + additional_investment

        profit = self._create_profit(
            ticker, unrealized_capital_gain=unrealized_capital_gain_new,
            reinvested_dividends=reinvested_dividends_new,
            profit_to_copy=investment_state_old.profit)

        investment_state = self._create_investment_state(
            ticker,
            quantity=quantity_new, total_cost=total_cost_new,
            average_cost=average_cost_new,
            total_commission=total_commission_new,
            asset_price=self.asset_price, total_invested=total_invested_new,
            profit=profit, investment_state_to_copy=investment_state_old)

        return investment_state


class RebalanceTransaction(Transaction):
    def __init__(self, investment, assets, asset_price, commission,
                 quantity=None, date=None):
        super().__init__("REBALANCE", investment, assets, quantity,
                         asset_price, commission, date)

        self.__value_by_ticker = None
        self.__rebalanced_value_by_ticker = None
        self.__difference_value_by_ticker = None
        self.__rebalanced_quantity_by_ticker = None
        self.__difference_quantity_by_ticker = None
        self.__total_cost_by_ticker = None
        self.__average_cost_by_ticker = None
        self.__total_commission_by_ticker = None
        self.__total_invested = None

        self.__realized_capital_gain_by_ticker = None
        self.__unrealized_capital_gain_by_ticker = None

    def update_investment_state(self, investment_state_old):

        self.__calculate_value_by_ticker(investment_state_old)

        self.__calculate_rebalanced_value_by_ticker(investment_state_old)
        self.__calculate_realized_capital_gain_by_ticker(investment_state_old)
        self.__calculate_cost_by_ticker(investment_state_old)
        self.__calculate_unrealized_capital_gain_by_ticker()
        self.__calculate_total_commission_by_ticker(investment_state_old)
        self.__calculate_total_invested_by_ticker(investment_state_old)

        cash_new = 0

        profit = Profit(self.investment, date=self.date,
                        realized_capital_gain_by_ticker=
                        self.__realized_capital_gain_by_ticker,
                        unrealized_capital_gain_by_ticker=
                        self.__unrealized_capital_gain_by_ticker,
                        profit_to_copy=investment_state_old.profit)

        investment_state = investment_module.InvestmentState(
            self.investment,
            quantity_by_ticker=self.__rebalanced_quantity_by_ticker,
            total_cost_by_ticker=self.__total_cost_by_ticker,
            average_cost_by_ticker=self.__average_cost_by_ticker,
            total_commission_by_ticker=self.__total_commission_by_ticker,
            asset_price_by_ticker=self.asset_price, cash=cash_new,
            total_invested=self.__total_invested,
            profit=profit,
            investment_state_to_copy=investment_state_old)

        return investment_state

    def __calculate_value_by_ticker(self, investment_state):
        self.__value_by_ticker = {}
        for ticker in self.ticker:
            quantity = investment_state.quantity_by_ticker[ticker]
            asset_price = self.asset_price[ticker]
            self.__value_by_ticker[ticker] = quantity * asset_price

    def __calculate_rebalanced_value_by_ticker(self, investment_state):

        total_value = sum(self.__value_by_ticker.values()) \
                      + investment_state.cash

        self.__rebalanced_value_by_ticker = {}
        self.__difference_value_by_ticker = {}
        self.__rebalanced_quantity_by_ticker = {}
        self.__difference_quantity_by_ticker = {}
        for ticker in self.ticker:
            weight = self.investment.weights[ticker]
            asset_price = self.asset_price[ticker]

            self.__rebalanced_value_by_ticker[ticker] = total_value * weight
            self.__rebalanced_quantity_by_ticker[ticker] = \
                self.__rebalanced_value_by_ticker[ticker] / asset_price
            self.__difference_value_by_ticker[ticker] = \
                self.__rebalanced_value_by_ticker[ticker] \
                - self.__value_by_ticker[ticker]
            self.__difference_quantity_by_ticker[ticker] = \
                self.__difference_value_by_ticker[ticker] / asset_price

    def __calculate_realized_capital_gain_by_ticker(self, investment_state):

        self.__realized_capital_gain_by_ticker = {}
        for ticker, diff_value in self.__difference_value_by_ticker.items():
            quantity = self.__difference_quantity_by_ticker[ticker]
            asset_price = self.asset_price[ticker]
            if diff_value > 0:
                # Buy
                self.__realized_capital_gain_by_ticker[ticker] = \
                    investment_state.profit.realized_capital_gain_by_ticker[ticker]
            else:
                # Sell
                add_realized_capital_gain = - quantity * (
                        asset_price
                        - investment_state.average_cost_by_ticker[ticker])
                self.__realized_capital_gain_by_ticker[ticker] = \
                    investment_state.profit.realized_capital_gain_by_ticker[
                        ticker] + add_realized_capital_gain

    def __calculate_cost_by_ticker(self, investment_state):
        self.__total_cost_by_ticker = {}
        self.__average_cost_by_ticker = {}
        for ticker, diff_value in self.__difference_value_by_ticker.items():
            if diff_value > 0:
                # Buy
                self.__total_cost_by_ticker[ticker] = \
                    investment_state.total_cost_by_ticker[ticker] + diff_value
            else:
                # Sell
                self.__total_cost_by_ticker[ticker] = \
                    investment_state.total_cost_by_ticker[ticker] \
                    + self.__difference_quantity_by_ticker[ticker] \
                    * investment_state.average_cost_by_ticker[ticker]

            self.__average_cost_by_ticker[ticker] = \
                self.__total_cost_by_ticker[ticker] \
                / self.__rebalanced_quantity_by_ticker[ticker]

    def __calculate_unrealized_capital_gain_by_ticker(self):

        self.__unrealized_capital_gain_by_ticker = {}
        for ticker, diff_value in self.__difference_value_by_ticker.items():
            quantity = self.__rebalanced_quantity_by_ticker[ticker]
            asset_price = self.asset_price[ticker]

            self.__unrealized_capital_gain_by_ticker[ticker] = \
                quantity * (asset_price
                            - self.__average_cost_by_ticker[ticker])

    def __calculate_total_commission_by_ticker(self, investment_state):

        self.__total_commission_by_ticker = {}
        for ticker, diff_value in self.__difference_value_by_ticker.items():
            self.__total_commission_by_ticker[ticker] = \
                investment_state.total_commission_by_ticker[
                    ticker] + self.commission

    def __calculate_total_invested_by_ticker(self, investment_state):

        self.__total_invested = investment_state.total_invested
        for ticker, diff_value in self.__difference_value_by_ticker.items():
            asset_price = self.asset_price[ticker]
            if diff_value > 0:
                # Buy
                additional_investment = \
                    self.__difference_quantity_by_ticker[ticker] \
                    * asset_price + self.commission
                self.__total_invested += additional_investment



from pyvest.data_reader.returns_data_reader import ReturnsDataReader
from pyvest.data_reader.factors_data_reader import FactorsDataReader

import pandas_datareader.data as web
from pandas_datareader.famafrench import get_available_datasets

# TODO: To be removed if/when pandas_datareader is updated.
import warnings
warnings.simplefilter(action="ignore", category=FutureWarning)


class FamaFrenchDataReader(FactorsDataReader, ReturnsDataReader):
    def __init__(self):
        super().__init__()

        self.__define_datasets()
        self.__define_portfolios()
        self.__define_renamed_portfolios()

    def read_factors(self, start_date, end_date, factors=None):
        # Description: returns the monthly Fama-French factors from Kenneth
        # French's website.
        # Inputs:
        #   -start_date: start date of the factors timeseries
        #   -end_date: end date of the factors timeseries
        #   -factor: list of strings representing the factors that should be
        #   included in the output DataFrame. By default, returns all the
        #   available factors.
        # Output:
        #   -factors_df: a Dataframe that contains the monthly factors

        data_factors_dict = web.DataReader('F-F_Research_Data_Factors',
                                           'famafrench', start=start_date,
                                           end=end_date)

        if factors is None:
            factors = list(data_factors_dict[0].columns)

        factors_df = data_factors_dict[0][factors]

        return factors_df

    def read_returns(self, data_set, start_date, end_date, freq="monthly",
                     weighting="value", portfolios=None, decile_only=True,
                     rename_portfolios=True):

        portfolios_dict_key = self.__get_portfolios_dict_key(freq, weighting)

        portfolios_dict = web.DataReader(data_set, 'famafrench',
                                         start=start_date, end=end_date)

        all_portfolios_df = portfolios_dict[portfolios_dict_key]

        if portfolios is not None:
            portfolios_df = all_portfolios_df[portfolios]
        else:
            portfolios = self.get_portfolio_names(data_set, decile_only)
            if portfolios is not None:
                portfolios_df = all_portfolios_df[portfolios]
            else:
                portfolios_df = all_portfolios_df

        if rename_portfolios:
            portfolios_df.columns = self.__rename_portfolios(data_set)

        return portfolios_df

    def get_available_datasets(self):
        return get_available_datasets()

    def get_portfolio_names(self, data_set, decile_only=False):

        portfolios = None
        if data_set in self.__quintile_decile_data_sets and not decile_only:
            portfolios = self.__quintile_portfolios \
                         + self.__decile_portfolios
        elif data_set in self.__quintile_decile_data_sets and decile_only:
            portfolios = self.__decile_portfolios
        # elif data_set in self.__five_by_five_data_sets:
        #     portfolios = self.__five_by_five_portfolios

        return portfolios

    ################################ PRIVATE ##################################

    def __get_portfolios_dict_key(self, freq, weighting):
        portfolios_dict_key = None
        if freq == "monthly" and weighting == "value":
            portfolios_dict_key = 0
        elif freq == "monthly" and weighting == "equal":
            portfolios_dict_key = 1
        elif freq == "annual" and weighting == "value":
            portfolios_dict_key = 2
        elif freq == "annual" and weighting == "equal":
            portfolios_dict_key = 3

        return portfolios_dict_key

    def __get_default_portfolios(self, data_set):

        portfolios = None
        if data_set in self.__quintile_decile_data_sets:
            portfolios = self.__decile_portfolios
        elif data_set in self.__five_by_five_data_sets:
            portfolios = ['SMALL LoBETA', 'ME1 BETA2', 'ME1 BETA3',
                          'ME1 BETA4', 'SMALL HiBETA', 'ME2 BETA1',
                          'ME2 BETA2', 'ME2 BETA3', 'ME2 BETA4',
                          'ME2 BETA5', 'ME3 BETA1', 'ME3 BETA2',
                          'ME3 BETA3', 'ME3 BETA4', 'ME3 BETA5',
                          'ME4 BETA1', 'ME4 BETA2', 'ME4 BETA3',
                          'ME4 BETA4', 'ME4 BETA5', 'BIG LoBETA',
                          'ME5 BETA2', 'ME5 BETA3', 'ME5 BETA4',
                          'BIG HiBETA']

        return portfolios

    def __rename_portfolios(self, data_set):
        # Description: rename the columns of "portfolios_data_set" for
        # better readability in plots
        # Input:
        #   -portfolios_data_set: the name of a dataset on Kenneth
        #   French's website
        # Output: a list of portfolio names

        if data_set in self.__quintile_decile_data_sets \
                or data_set in self.__prior_data_sets:
            new_portfolio_names = ['1', '2', '3', '4', '5', '6', '7', '8',
                                   '9', '10']
        # elif data_set in self.__five_by_five_data_sets:
        else:
            new_portfolio_names = ['(' + str(i) + ',' + str(j) + ')'
                                   for i in range(1, 6) for j in range(1, 6)]

        return new_portfolio_names

    def __define_datasets(self):
        self.__quintile_decile_data_sets = ['Portfolios_Formed_on_BETA',
                                            'Portfolios_Formed_on_ME',
                                            'Portfolios_Formed_on_BE-ME']

        self.__prior_data_sets = ['10_Portfolios_Prior_12_2',
                                  '10_Portfolios_Prior_1_0',
                                  '10_Portfolios_Prior_60_13']

        self.__five_by_five_data_sets = ['25_Portfolios_ME_BETA_5x5',
                                         '25_Portfolios_5x5',
                                         '25_Portfolios_ME_Prior_12_2',
                                         '25_Portfolios_ME_Prior_1_0',
                                         '25_Portfolios_ME_Prior_60_13']

    def __define_portfolios(self):

        self.__quintile_portfolios = [
            'Lo 20', 'Qnt 2', 'Qnt 3', 'Qnt 4', 'Hi 20'
        ]

        self.__decile_portfolios = [
            'Lo 10', 'Dec 2', 'Dec 3', 'Dec 4', 'Dec 5', 'Dec 6', 'Dec 7',
            'Dec 8', 'Dec 9', 'Hi 10'
        ]

        self.__prior_portfolios = [
            'Lo PRIOR', 'PRIOR 2', 'PRIOR 3', 'PRIOR 4', 'PRIOR 5', 'PRIOR 6',
            'PRIOR 7', 'PRIOR 8', 'PRIOR 9', 'Hi PRIOR'
        ]

        # TODO: NOT GENERAL ENOUGH
        self.__five_by_five_portfolios = [
            'SMALL LoBETA', 'ME1 BETA2', 'ME1 BETA3', 'ME1 BETA4',
            'SMALL HiBETA', 'ME2 BETA1', 'ME2 BETA2', 'ME2 BETA3', 'ME2 BETA4',
            'ME2 BETA5', 'ME3 BETA1', 'ME3 BETA2', 'ME3 BETA3', 'ME3 BETA4',
            'ME3 BETA5', 'ME4 BETA1', 'ME4 BETA2', 'ME4 BETA3', 'ME4 BETA4',
            'ME4 BETA5', 'BIG LoBETA', 'ME5 BETA2', 'ME5 BETA3', 'ME5 BETA4',
            'BIG HiBETA'
        ]

    def __define_renamed_portfolios(self):

        self.renamed_portfolios_dict = {
            'Lo 20': "Q1",
            'Qnt 2': "Q2",
            'Qnt 3': "Q3",
            'Qnt 4': "Q4",
            'Hi 20': "Q5",
            'Lo 10': "1",
            'Dec 2': "2",
            'Dec 3': "3",
            'Dec 4': "4",
            'Dec 5': "5",
            'Dec 6': "6",
            'Dec 7': "7",
            'Dec 8': "8",
            'Dec 9': "9",
            'Hi 10': "10"
        }



import pyvest.investment_universe.investment_universe as inv_univ
from pyvest.general.portfolio import Portfolio
from pyvest.general.general import standard_utility_function

import numpy as np

from scipy.optimize import minimize
from scipy.optimize import LinearConstraint
from scipy.optimize import Bounds


class Investor:

    def __init__(self, investment_universe, gamma, wealth=None, portfolio=None,
                 utility_function=None, name=None,
                 optimization_tolerance=1e-6):

        self.__assign_investment_universe(investment_universe)
        self.__assign_gamma(gamma)
        self.__assign_wealth(wealth)
        self.__assign_portfolio(portfolio)
        self.__assign_name(name)
        self.__assign_utility_function(utility_function)
        self.__assign_optimization_tolerance(optimization_tolerance)

        self.__optimal_portfolio = None

    def __repr__(self):
        return self.__generate_output()

    def __str__(self):
        return self.__generate_output()

    ################################# ATTRIBUTES ##############################

    @property
    def investment_universe(self):
        return self.__investment_universe

    @investment_universe.setter
    def investment_universe(self, value):
        self.__assign_investment_universe(value)

    @property
    def utility_function(self):
        return self.__utility_function

    @utility_function.setter
    def utility_function(self, value):
        self.__assign_utility_function(value)

    @property
    def gamma(self):
        return self.__gamma

    @gamma.setter
    def gamma(self, value):
        self.__assign_gamma(value)

    @property
    def wealth(self):
        return self.__wealth

    @wealth.setter
    def wealth(self, value):
        self.__assign_wealth(value)

    @property
    def portfolio(self):
        return self.__portfolio

    @portfolio.setter
    def portfolio(self, value):
        self.__assign_portfolio(value)

    @property
    def optimal_portfolio(self):
        return self.__optimal_portfolio

    @property
    def portfolio_utility(self):
        return self.__utility_function(self.__portfolio.weights,
                                       self.__portfolio.augmented_mu,
                                       self.__portfolio.augmented_cov,
                                       self.gamma)

    @property
    def optimal_portfolio_utility(self):
        return self.__utility_function(self.__optimal_portfolio.weights,
                                       self.__optimal_portfolio.augmented_mu,
                                       self.__optimal_portfolio.augmented_cov,
                                       self.gamma)

    @property
    def name(self):
        return self.__name

    ################################# PUBLIC ##################################

    def calculate_optimal_portfolio(self, x0=None):

        nb_risky_assets = len(self.__investment_universe.mu)
        nb_assets = nb_risky_assets if self.__investment_universe.r_f is None \
            else nb_risky_assets + 1

        # Initial guess (seed value)
        if x0 is None:
            x0 = np.ones(nb_assets) / nb_assets

        # Sum portfolio weights equals 1 constraint
        sum_weights_assets_equals_one_constraint = LinearConstraint(
            np.ones(nb_assets), 1, 1)

        if self.__investment_universe.min_weight is None \
                and self.__investment_universe.min_weight_r_f is None:
            min_weights = None
        elif self.__investment_universe.r_f is None:
            min_weights = self.__investment_universe.min_weight
        else:
            min_weights = np.append(self.__investment_universe.min_weight,
                                    self.__investment_universe.min_weight_r_f)

        min_weights_bound = Bounds(min_weights, np.inf) \
            if min_weights is not None else None

        optimal_portfolio_result = minimize(
            lambda x: - self.__utility_function(
                x, self.__investment_universe.augmented_mu,
                self.__investment_universe.augmented_cov, self.__gamma), x0,
            bounds=min_weights_bound,
            constraints=[sum_weights_assets_equals_one_constraint],
            tol=self.__optimization_tolerance)

        # Assign results
        self.__optimal_portfolio = Portfolio(
            optimal_portfolio_result.x, self.__investment_universe.mu,
            self.__investment_universe.cov, r_f=self.__investment_universe.r_f)

        return self.__optimal_portfolio

    def calculate_indifference_curve(self, utility, min_std=0, max_std=10,
                                     nb_points=1000):
        # TODO: MAKE IT WORK FOR A GENERAL UTILITY FUNCTION

        delta = (max_std - min_std) / (nb_points - 1)

        std_array = np.arange(min_std, max_std + delta, delta)
        mu_array = utility + self.gamma * std_array ** 2

        return std_array, mu_array

    def calculate_portfolio_utility(self, portfolio):

        if isinstance(portfolio, Portfolio):
            portfolio_obj = portfolio
        elif (isinstance(portfolio, list) or isinstance(portfolio, np.ndarray)
              or isinstance(portfolio, tuple)):
            portfolio_obj = Portfolio(portfolio, self.investment_universe.mu,
                                      self.investment_universe.cov,
                                      r_f=self.investment_universe.r_f,
                                      assets=self.investment_universe.assets)
        else:
            raise TypeError("The variable 'portfolio' must be an object of "
                            "type Portfolio or a list of weights.")

        return self.__utility_function(portfolio_obj.weights,
                                       portfolio_obj.augmented_mu,
                                       portfolio_obj.augmented_cov, self.gamma)

    ################################ PRIVATE ##################################

    def __assign_investment_universe(self, investment_universe):
        self.__investment_universe = investment_universe
        if isinstance(investment_universe, inv_univ.InvestmentUniverse):
            self.__investment_universe = investment_universe
        else:
            raise TypeError("The parameter 'investment_universe' must be an "
                            "instance of InvestmentUniverse.")

    def __assign_utility_function(self, utility_function):
        if utility_function is None:
            self.__utility_function = standard_utility_function
        elif callable(utility_function):
            self.__utility_function = utility_function
        else:
            raise TypeError("The parameter 'utility_function' must be "
                            "callable.")

    def __assign_gamma(self, gamma):
        if type(gamma) is float or type(gamma) is int:
            self.__gamma = float(gamma)
        else:
            raise TypeError("The parameter 'gamma' must of type float or int.")

    def __assign_wealth(self, wealth):
        if wealth is None:
            self.__wealth = None
        elif type(wealth) is float or type(wealth) is int:
            self.__wealth = float(wealth)
        else:
            raise TypeError("The parameter 'wealth' must be None or of type "
                            "float or int.")

    def __assign_portfolio(self, portfolio):
        if portfolio is None or type(portfolio) is Portfolio:
            self.__portfolio = portfolio
        elif type(portfolio) is list \
                and self.__investment_universe is not None:
            self.__portfolio = Portfolio(
                portfolio, self.__investment_universe.mu,
                self.__investment_universe.cov,
                r_f=self.__investment_universe.r_f,
                assets=self.__investment_universe.assets)
        else:
            raise TypeError("The parameter 'portfolio' must be None or of "
                            "type Portfolio or list.")

    def __assign_name(self, name):
        if name is None or type(name) is str:
            self.__name = name
        else:
            raise TypeError("The parameter 'name' must be None or of type "
                            "str.")

    def __assign_optimization_tolerance(self, optimization_tolerance):
        if type(optimization_tolerance) is float:
            self.__optimization_tolerance = optimization_tolerance
        else:
            raise TypeError("The parameter 'optimization_tolerance' must be "
                            "of type float.")

    def __generate_output(self):

        if self.name is not None:
            output = "name: {}\n".format(self.name)
        else:
            output = ""

        output += "gamma: {}".format(str(self.gamma))

        if self.wealth is not None:
            output += "\nwealth: {}".format(str(self.wealth))

        if self.portfolio is not None:
            output += "\nportfolio:"
            for portfolio_line in str(self.portfolio).splitlines():
                output += "\n  -{}".format(portfolio_line)

            output += "\n  -{}: {}".format("utility", self.portfolio_utility)

        if self.optimal_portfolio is not None:
            output += "\noptimal portfolio:"
            for portfolio_line in str(self.optimal_portfolio).splitlines():
                output += "\n  -{}".format(portfolio_line)

            output += "\n  -{}: {}".format("utility",
                                           self.optimal_portfolio_utility)

        return output



import pandas as pd
import datetime

import pyvest.investment as investment_module
from pyvest.investment.transaction import BuyTransaction, SellTransaction, \
    DividendTransaction, ReinvestedDividendTransaction, RebalanceTransaction


class Investment:
    def __init__(self, tickers, cash, weights=None, start_date=None,
                 end_date=None, frequency=None, dividend_commission=0,
                 rebalance_commission=0, reinvest_dividends=True,
                 rebalance=False, from_historical_data=True,
                 data_reader=None):

        self.__tickers = [tickers] if isinstance(tickers, str) else tickers

        self.__cash = cash

        self.__initialize_frequency(frequency)

        self.__start_date = self.__preprocess_date(start_date)
        self.__end_date = self.__preprocess_date(end_date)
        self.__assets = {}
        self.__asset_types = {}

        self.__weights = weights
        self.__reinvest_dividends = reinvest_dividends
        self.__rebalance = rebalance

        self.__data_reader = data_reader

        self.__unprocessed_transactions = {}
        self.__initialize_state_history()

        if from_historical_data and self.__start_date is not None \
                and self.__end_date is not None:
            self.__create_assets_from_historical_data()
            self.__initialize_dates()
            self.__create_dividend_transactions(dividend_commission,
                                                reinvest_dividends)
            if weights is not None:
                self.rebalance(self.__start_date, rebalance_commission)
            if weights is not None and rebalance:
                self.__create_rebalance_transactions(rebalance_commission)
        else:
            raise ValueError("from_historical_data must be True and "
                             "start_date and end_date must be provided.")

        self.__visualizer = None

    def __repr__(self):
        return self.__generate_output()

    def __str__(self):
        return self.__generate_output()

    @property
    def current_state(self):
        return self.__current_state

    @property
    def assets(self):
        return self.__assets

    @property
    def weights(self):
        return self.__weights

    @property
    def transactions(self):
        return self.__transactions_history

    @property
    def history(self):
        return self.__state_history

    @property
    def reinvest_dividends(self):
        return self.__reinvest_dividends

    @property
    def start_date(self):
        return self.__start_date

    @property
    def end_date(self):
        return self.__end_date

    @property
    def visualizer(self):
        return self.__visualizer

    def get_most_recent_state(self, date):

        most_recent_state = None

        for state_date, states_list in self.__state_history.items():
            if state_date > date:
                break
            most_recent_state = states_list[-1]

        return most_recent_state

    def buy(self, ticker, date, quantity, price=None, commission=0):

        date = self.__preprocess_date(date)

        if price is not None:
            buy_transaction = BuyTransaction(self, self.__assets[ticker],
                                             quantity, price, commission, date)
        elif date is not None:
            price = self.__get_asset_price(ticker, date)
            buy_transaction = BuyTransaction(self, self.__assets[ticker],
                                             quantity, price, commission, date)
        else:
            raise ValueError("Either both 'quantity' and 'price' or 'date' "
                             "must be different from None.")

        self.__process_transaction(date, buy_transaction)

        return buy_transaction

    def sell(self, ticker, date, quantity, price=None, commission=0):

        date = self.__preprocess_date(date)

        if price is not None:
            sell_transaction = SellTransaction(self, self.__assets[ticker],
                                               quantity, price, commission,
                                               date)
        elif date is not None:
            price = self.__get_asset_price(ticker, date)
            sell_transaction = SellTransaction(self, self.__assets[ticker],
                                               quantity, price, commission,
                                               date)
        else:
            raise ValueError("Either both 'quantity' and 'price' or 'date' "
                             "must be different from None.")

        self.__process_transaction(date, sell_transaction)

        return sell_transaction

    def add_dividend(self, ticker, date, dividend_by_share, commission=0,
                     reinvest_dividends=True):

        date = self.__preprocess_date(date)

        asset_price = self.__get_asset_price(ticker, date)

        if reinvest_dividends:
            dividend_transaction = ReinvestedDividendTransaction(
                self, self.__assets[ticker], asset_price, commission,
                dividend_by_share, date=date)
        else:
            dividend_transaction = DividendTransaction(
                self, self.__assets[ticker], asset_price, commission,
                dividend_by_share, date=date)

        self.__process_transaction(date, dividend_transaction)

        return dividend_transaction

    def rebalance(self, date, commission=0):
        date = self.__preprocess_date(date)

        asset_price_by_ticker = {}
        for ticker in self.__tickers:
            asset_price = self.__get_asset_price(ticker, date)
            asset_price_by_ticker[ticker] = asset_price

        rebalance_transaction = RebalanceTransaction(
            self, self.__assets, asset_price_by_ticker, commission, date=date)

        self.__process_transaction(date, rebalance_transaction)

        return rebalance_transaction

    def calculate_transactions(self):

        self.__initialize_state_history()

        unprocessed_trans_dates = \
            self.__get_sorted_dates_from_map(self.__unprocessed_transactions)

        for trans_date in unprocessed_trans_dates:
            transactions_list = self.__unprocessed_transactions[trans_date]

            if trans_date not in self.__state_history:
                self.__state_history[trans_date] = []

            if trans_date not in self.__transactions_history:
                self.__transactions_history[trans_date] = []

            for transaction in transactions_list:
                new_state = transaction.update_investment_state(
                    self.__current_state)

                self.__state_history[trans_date].append(new_state)
                self.__current_state = new_state

                self.__transactions_history[trans_date].append(transaction)

        end_state = self.__calculate_end_state()
        self.__state_history[self.__end_date] = [end_state]

    def calculate_profit(self, date, asset_prices=None):

        if type(date) == str:
            date = self.__preprocess_date(date)

        if asset_prices is None:
            asset_prices = {}
            for ticker in self.__tickers:
                asset_prices[ticker] = self.__get_asset_price(ticker, date)

        state = self.__get_most_recent_state(date)

        # Update capital gains
        unrealized_capital_gain_by_ticker = {}
        for ticker in self.__tickers:
            unrealized_capital_gain = \
                state.quantity_by_ticker[ticker] \
                * (asset_prices[ticker] - state.average_cost_by_ticker[ticker])
            unrealized_capital_gain_by_ticker[ticker] = unrealized_capital_gain

        profit = investment_module.Profit(
            self,
            date=date,
            unrealized_capital_gain_by_ticker=
            unrealized_capital_gain_by_ticker,
            profit_to_copy=state.profit)

        return profit

    def calculate_profit_range(self, start_date=None, end_date=None):

        if start_date is None:
            start_date = self.__start_date

        if end_date is None:
            end_date = self.__end_date

        processed_start_date = pd.to_datetime(start_date).tz_localize(None)
        processed_end_date = pd.to_datetime(end_date).tz_localize(None)

        profits_list = []
        for date in self.__history_dates:
            processed_date = pd.to_datetime(date).tz_localize(None)
            if processed_start_date <= processed_date <= processed_end_date:
                profit = self.calculate_profit(date=processed_date)
                profits_list.append(profit)

        return profits_list

    def get_transactions(self, transaction_type=None, start_date=None,
                         end_date=None):

        processed_start_date = pd.to_datetime(start_date).tz_localize(None) \
            if start_date is not None else None
        processed_end_date = pd.to_datetime(end_date).tz_localize(None) \
            if end_date is not None else None

        transactions_sub_dict = {}
        for date, transactions_list in self.__transactions_history.items():
            if (start_date is None or date >= processed_start_date) \
                    and (end_date is None or date <= processed_end_date):
                for transaction in transactions_list:
                    if transaction_type is None \
                            or transaction.type == transaction_type:
                        if date not in transactions_sub_dict:
                            transactions_sub_dict[date] = []
                        transactions_sub_dict[date].append(transaction)

        return transactions_sub_dict

    def calculate_return_range(self, start_date=None, end_date=None,
                               percentage=True):

        if start_date is None:
            start_date = self.__start_date

        if end_date is None:
            end_date = self.__end_date

        processed_start_date = pd.to_datetime(start_date).tz_localize(None)
        processed_end_date = pd.to_datetime(end_date).tz_localize(None)

        returns_list = []
        for date in self.__history_dates:
            processed_date = pd.to_datetime(date).tz_localize(None)
            if processed_start_date <= processed_date <= processed_end_date:
                inv_return = self.__calculate_return_on_date(
                    processed_date, percentage=percentage)
                returns_list.append(inv_return)

        return returns_list

    def get_state(self, date):

        processed_date = pd.to_datetime(date).tz_localize(None)

        most_recent_state = self.__get_most_recent_state(processed_date)
        profit = self.calculate_profit(processed_date)

        asset_price_by_ticker = {}
        for ticker in self.__tickers:
            asset_price_by_ticker[ticker] = self.__get_asset_price(
                ticker, processed_date)

        state = investment_module.InvestmentState(
            self, asset_price_by_ticker=asset_price_by_ticker, profit=profit,
            investment_state_to_copy=most_recent_state)

        return state

    def calculate_weight(self, date):

        processed_date = pd.to_datetime(date).tz_localize(None)

        state = self.get_state(date)

        weight_by_ticker = state.weight_by_ticker

        dated_weight_by_ticker = {}
        for ticker, weight in weight_by_ticker.items():
            dated_weight = \
                investment_module.investment_dated_info.InvestmentDatedInfo(
                    processed_date, ticker=ticker,  weight=weight)
            dated_weight_by_ticker[ticker] = dated_weight

        return dated_weight_by_ticker

    def calculate_weight_range(self, start_date=None, end_date=None):

        if start_date is None:
            start_date = self.__start_date

        if end_date is None:
            end_date = self.__end_date

        processed_start_date = pd.to_datetime(start_date).tz_localize(None)
        processed_end_date = pd.to_datetime(end_date).tz_localize(None)

        dated_weights_list = []
        for date in self.__history_dates:
            processed_date = pd.to_datetime(date).tz_localize(None)
            if processed_start_date <= processed_date <= processed_end_date:
                dated_weight_by_ticker = self.calculate_weight(processed_date)
                dated_weights_list.append(dated_weight_by_ticker)

        return dated_weights_list

    def calculate_value(self, date):

        processed_date = pd.to_datetime(date).tz_localize(None)

        state = self.get_state(date)

        value_by_ticker = state.value_by_ticker

        dated_value_by_ticker = {}
        for ticker, value in value_by_ticker.items():
            dated_value = \
                investment_module.investment_dated_info.InvestmentDatedInfo(
                    processed_date, ticker=ticker,  value=value)
            dated_value_by_ticker[ticker] = dated_value

        return dated_value_by_ticker

    def calculate_value_range(self, start_date=None, end_date=None):

        if start_date is None:
            start_date = self.__start_date

        if end_date is None:
            end_date = self.__end_date

        processed_start_date = pd.to_datetime(start_date).tz_localize(None)
        processed_end_date = pd.to_datetime(end_date).tz_localize(None)

        dated_values_list = []
        for date in self.__history_dates:
            processed_date = pd.to_datetime(date).tz_localize(None)
            if processed_start_date <= processed_date <= processed_end_date:
                dated_value_by_ticker = self.calculate_value(processed_date)
                dated_values_list.append(dated_value_by_ticker)

        return dated_values_list

    ############################## PLOT METHODS ###############################

    def plot_profit(self, compare_with=None,
                    start_date=None, end_date=None, figsize=(16, 9),
                    legend='best', show_transactions=False):

        self.__prepare_visualizer(compare_with)
        self.__visualizer.plot_profit(start_date, end_date, figsize, legend,
                                      show_transactions)

    def plot_return(self, compare_with=None,
                    start_date=None, end_date=None, figsize=(16, 9),
                    legend='best', show_transactions=False):
        self.__prepare_visualizer(compare_with)
        self.__visualizer.plot_return(start_date, end_date, figsize, legend,
                                      show_transactions)

    def plot_weight(self, compare_with=None,
                    start_date=None, end_date=None, figsize=(16, 9),
                    legend='best', show_transactions=False):
        self.__prepare_visualizer(compare_with)
        self.__visualizer.plot_weight(start_date, end_date, figsize, legend,
                                      show_transactions)

    def plot_value(self, compare_with=None,
                   start_date=None, end_date=None, figsize=(16, 9),
                   legend='best', show_transactions=False):
        self.__prepare_visualizer(compare_with)
        self.__visualizer.plot_value(start_date, end_date, figsize, legend,
                                      show_transactions)

    ################################ PRIVATE ##################################

    def __initialize_frequency(self, frequency):

        is_portfolio = False
        for ticker in self.__tickers:
            if self.__is_portfolio(ticker):
                is_portfolio = True

        if is_portfolio and (frequency is None or frequency == "1mo"):
            # Portfolios are only available at a monthly frequency.
            self.__frequency = "1mo"
        elif is_portfolio:
            raise ValueError("Fama-French portfolios are only available at a "
                             "monthly frequency.")
        elif frequency is None:
            self.__frequency = "1d"
        else:
            self.__frequency = frequency

    def __initialize_state_history(self):
        self.__transactions_history = {}
        self.__current_state = investment_module.InvestmentState(
            self, cash=self.__cash)
        self.__state_history = {self.__start_date: [self.__current_state]}

    def __preprocess_date(self, date):

        if isinstance(date, str):
            full_date = self.__preprocess_string_date(date)
        else:
            full_date = date

        preprocessed_date = pd.to_datetime(full_date).tz_localize(None)

        return preprocessed_date

    def __preprocess_string_date(self, date_str):

        full_date_str = date_str
        if self.__frequency == "1mo":
            if len(date_str) == 7 and date_str[4] == "-":
                full_date_str = date_str + "-01"
            else:
                raise ValueError("Monthly dates must be in the format YYYY-MM")

        return full_date_str

    def __process_transaction(self, date, transaction):

        if date in self.__unprocessed_transactions:
            self.__unprocessed_transactions[date].append(transaction)
        else:
            self.__unprocessed_transactions[date] = [transaction]

        self.calculate_transactions()

    def __get_sorted_dates_from_map(self, dates_map):
        return sorted(list(dates_map.keys()))

    def __create_assets_from_historical_data(self):

        for ticker in self.__tickers:
            if self.__is_portfolio(ticker):
                # Portfolio (FF)
                asset = investment_module.StockPortfolio(
                    ticker, start_date=self.__start_date,
                    end_date=self.__end_date, frequency=self.__frequency,
                    data_reader=self.__data_reader)
                asset_type = "portfolio"
            else:
                # Stock (yfinance)
                asset = investment_module.Stock(ticker,
                                                start_date=self.__start_date,
                                                end_date=self.__end_date,
                                                frequency=self.__frequency,
                                                data_reader=self.__data_reader)
                asset_type = "stock"

            self.__assets[ticker] = asset
            self.__asset_types[ticker] = asset_type

    def __create_dividend_transactions(self, commission, reinvest_dividends):
        for ticker, asset in self.__assets.items():
            if self.__asset_types[ticker] == "stock":
                for date, dividend_amount in asset.dividends.items():
                    self.add_dividend(ticker, date.tz_localize(None),
                                      dividend_amount,
                                      commission=commission,
                                      reinvest_dividends=reinvest_dividends)

    def __create_rebalance_transactions(self, commission):
        end_of_month_dates = self.__get_end_of_month_dates()
        for date in end_of_month_dates:
            self.rebalance(date, commission)

    def __get_end_of_month_dates(self):
        end_of_month_dates = []
        previous_date = None
        previous_date_month = None
        for date in self.__history_dates:
            current_date_month = date.month
            if previous_date_month is not None \
                    and current_date_month != previous_date_month:
                end_of_month_dates.append(previous_date)
            previous_date = date
            previous_date_month = current_date_month

        last_date = previous_date
        last_date_month = previous_date_month

        if self.__frequency == "1d":
            day_after_last_date = last_date + datetime.timedelta(days=1)
            day_after_last_date_month = day_after_last_date.month
            if day_after_last_date_month != last_date_month:
                end_of_month_dates.append(last_date)
        elif self.__frequency == "1mo" or self.__frequency == "3mo":
            end_of_month_dates.append(last_date)

        return end_of_month_dates

    def __get_asset_price(self, ticker, date, column="Close"):
        asset_price = self.__assets[ticker].get_price(date)

        return asset_price

    def __get_most_recent_state(self, date):

        most_recent_state = None

        most_recent_date = None
        state_dates = self.__get_sorted_dates_from_map(self.__state_history)
        for state_date in state_dates:
            if date < state_date:
                break
            most_recent_date = state_date

        if most_recent_date is not None:
            most_recent_state = self.__state_history[most_recent_date][-1]

        return most_recent_state

    def __calculate_end_state(self):

        profit = self.calculate_profit(self.__end_date)
        most_recent_state = self.__get_most_recent_state(self.__end_date)

        end_state = investment_module.InvestmentState(
            self, profit=profit, investment_state_to_copy=most_recent_state)

        return end_state

    def __initialize_dates(self):
        self.__start_date = self.__preprocess_date(
            self.__assets[self.__tickers[0]].start_date)
        self.__end_date = self.__preprocess_date(
            self.__assets[self.__tickers[0]].end_date)
        self.__history_dates = self.__assets[self.__tickers[0]].dates

    def __calculate_return_on_date(self, date, percentage=True):

        profit = self.calculate_profit(date)

        total_profit = \
            profit.non_reinvested_dividends + profit.realized_capital_gain \
            + profit.reinvested_dividends + profit.unrealized_capital_gain

        total_return_value = total_profit / self.current_state.total_invested

        if percentage:
            total_return_value *= 100

        total_return = \
            investment_module.investment_dated_info.InvestmentDatedInfo(
            profit.date, total_return=total_return_value)

        return total_return

    def __generate_output(self):
        output = str(self.__current_state)

        return output

    def __is_portfolio(self, ticker):

        is_portfolio = False

        factor_tickers = ["Mkt-RF", "SMB", "HML", "RF", "Mkt"]
        portfolio_tickers = ["beta_" + str(i) for i in range(1, 11)] + \
                            ["size_" + str(i) for i in range(1, 11)] + \
                            ["value_" + str(i) for i in range(1, 11)] + \
                            ["momentum_" + str(i) for i in range(1, 11)]

        if ticker in factor_tickers or ticker in portfolio_tickers:
            is_portfolio = True

        return is_portfolio

    def __prepare_visualizer(self, compare_with):
        investment = [self]
        if isinstance(compare_with, Investment):
            investment.append(compare_with)
        elif isinstance(compare_with, list):
            investment.extend(compare_with)

        if self.__visualizer is None:
            self.__visualizer = \
                investment_module.InvestmentVisualizer(investment)
        else:
            self.__visualizer.investments = investment



class InvestmentDatedInfo:
    def __init__(self, date, ticker=None, total_return=None, weight=None,
                 value=None):

        self.__date = date
        self.__ticker = ticker
        self.__total_return = total_return
        self.__weight = weight
        self.__value = value

    def __repr__(self):
        return self.__generate_output()

    def __str__(self):
        return self.__generate_output()

    @property
    def date(self):
        return self.__date

    @property
    def ticker(self):
        return self.__ticker

    @property
    def total_return(self):
        return self.__total_return

    @property
    def weight(self):
        return self.__weight

    @property
    def value(self):
        return self.__value

    def __generate_output(self):
        output = ""

        if self.__date is not None:
            output += "Date: " + str(self.__date)

        if self.__total_return is not None:
            if len(output) > 0:
                output += "\n"
            output += "Total Return: " \
                      + str(self.__total_return)

        if self.__weight is not None:
            if len(output) > 0:
                output += "\n"
            output += "Weight: " \
                      + str(self.__weight)

        if self.__value is not None:
            if len(output) > 0:
                output += "\n"
            output += "Value: " \
                      + str(self.__value)

        return output



from pyvest.investment_universe.investment_universe import *
from pyvest.investment_universe.investment_universe_visualizer import *
from pyvest.investment_universe.investor import *

__all__ = ['InvestmentUniverse', 'InvestmentUniverseVisualizer', 'Investor']



import numpy as np
import statsmodels.api as sm

from pyvest.factor_model.factor_model_visualizer import FactorModelVisualizer
from pyvest.factor_model.regression import Regression


class FactorModel:

    def __init__(self, r_f, factors, returns, name=None):

        if self.__check_r_f(r_f):
            self.__r_f = r_f
        if self.__check_factors(factors):
            self.__factors = factors
        if self.__check_returns(returns):
            self.__returns = returns

        self.__X = self.__factors
        self.__X = sm.add_constant(self.__X)
        self.__Y = self.__returns.sub(self.__r_f, axis=0)
        self.__name = name

        self.__regressions_dict = None
        self.__construct_regressions()
        self.__realized_average_returns = None
        self.__predicted_average_returns = None

        self.__portfolios = self.__Y.columns

        self.__visualizer = None

    def __getitem__(self, key):
        regressions = None
        if self.__regressions_dict is not None:
            regressions = self.__regressions_dict[key]

        return regressions

    def __iter__(self):
        return iter(self.__regressions_dict)

    def keys(self):
        return self.__regressions_dict.keys()

    def items(self):
        return self.__regressions_dict.items()

    def values(self):
        return self.__regressions_dict.values()

    ##################### X ###################    
    @property
    def X(self):
        return self.__X

    @X.setter
    def X(self, value):
        self.__X = value

        ##################### Y ###################

    @property
    def Y(self):
        return self.__Y

    @Y.setter
    def Y(self, value):
        self.__Y = value

    ##################### r_f ###################    
    @property
    def r_f(self):
        return self.__r_f

    ##################### regressions ###################
    @property
    def regressions(self):
        return self.__regressions_dict

    ##################### regressions_calculated ###################
    @property
    def regressions_calculated(self):
        ret = True if len(self.values()) > 0 else False
        for regression in self.values():
            if regression.regression_results is None:
                ret = False

        return ret

    ##################### realized_average_returns ###################    
    @property
    def realized_average_returns(self):
        return self.__realized_average_returns

    ##################### predicted_average_returns ###################    
    @property
    def predicted_average_returns(self):
        return self.__predicted_average_returns

    ##################### name ###################
    @property
    def name(self):
        return self.__name

    @name.setter
    def name(self, value):
        self.__name = value

    ##################### portfolios ###################
    @property
    def portfolios(self):
        return list(self.__Y.columns)

    ##################### visualizer ###################
    @property
    def visualizer(self):
        return self.__visualizer

    ########################## PUBLIC ##########################    

    def calculate_regressions(self, return_results=True):
        # Description: estimates CAPM regressions for all self.__Y in the
        # self.__Y DataFrame and returns dictionaries with estimated alphas,
        # estimated betas, and associated confidence interval of every
        # self.__Y.
        # Output: dictionaries containing the estimated alpha, beta, and
        # confidence interval of all self.__Y

        for regression in self.values():
            regression.calculate_regression(return_results=False)

        if return_results:
            return self.__regressions_dict

    def calculate_realized_vs_predicted_average_returns(self,
                                                        return_results=True):
        if not self.regressions_calculated:
            self.calculate_regressions(return_results=False)

        rf_mean = self.__r_f.mean()

        factors_mean = self.__factors.mean()
        if len(factors_mean.shape) == 0:
            factors_mean = np.array([factors_mean])

        beta_list = [rr.beta for rr
                     in self.__regressions_dict.values()]
        alpha_list = [rr.alpha for rr
                      in self.__regressions_dict.values()]

        self.__realized_average_returns = {}
        self.__predicted_average_returns = {}
        for ptf, i in zip(self.portfolios, range(len(self.portfolios))):
            betas_dot_factors_mean = np.dot(factors_mean, beta_list[i])
            realized_average_returns = \
                alpha_list[i] + rf_mean + betas_dot_factors_mean
            predicted_average_returns = rf_mean + betas_dot_factors_mean

            self.__realized_average_returns[ptf] = realized_average_returns
            self.__predicted_average_returns[ptf] = predicted_average_returns

        if return_results:
            return self.__realized_average_returns, \
                   self.__predicted_average_returns

    def plot(self, compare_with=None, labels=None, colors=None,
             error_bars_colors=None, portfolios=None, legend='upper left',
             min_return=0, max_return=1.5, confidence_level=None, sml=False,
             beta_min=0, beta_max=2):

        self.__construct_visualizer(compare_with=compare_with, labels=labels,
                                    colors=colors,
                                    error_bars_colors=error_bars_colors)

        if sml:
            self.visualizer.plot_sml(beta_min=beta_min, beta_max=beta_max,
                                     legend=legend, portfolios=portfolios,
                                     confidence_level=confidence_level)
        else:
            self.visualizer.plot_realized_vs_predicted_average_return(
                min_return=min_return, max_return=max_return, legend=legend,
                portfolios=portfolios, confidence_level=confidence_level)

    ########################## PRIVATE ##########################

    def __construct_visualizer(self, compare_with=None, labels=None,
                               colors=None, error_bars_colors=None):
        factor_models = [self]
        if isinstance(compare_with, FactorModel):
            factor_models.append(compare_with)
        elif isinstance(compare_with, list):
            factor_models.extend(compare_with)

        if labels is None:
            labels_iter = iter([1, 2, 3, 4])
            labels = [fm.name if fm.name is not None else next(labels_iter)
                      for fm in factor_models]

        self.__visualizer = FactorModelVisualizer(
            factor_models, labels=labels, colors=colors,
            error_bars_colors=error_bars_colors)

    def __construct_regressions(self):
        self.__regressions_dict = {}
        for column_name in self.__Y.columns:
            regression = Regression(self.__X, self.__Y[column_name],
                                    name=self.__name + " - " + column_name)
            self.__regressions_dict[column_name] = regression

    def __check_r_f(self, r_f):
        if len(r_f.shape) != 1 and (len(r_f.shape) != 2 or r_f.shape[1] != 1):
            raise ValueError("The argument of 'r_f' must be one-dimensional.")
        return True

    def __check_factors(self, factors):
        if len(factors) != len(self.__r_f):
            raise ValueError(
                "The arguments of 'factors' and 'rf' must be of the same length.")
        return True

    def __check_returns(self, returns):
        if len(returns) != len(self.__r_f):
            raise ValueError(
                "The arguments of 'returns' and 'rf' must be of the same length.")
        return True
